# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  sqs_init.py                                                                          #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

import time
import boto3
from urllib.parse import urlsplit
from botocore.config import Config as boto3config
from botocore.exceptions import ClientError, EndpointConnectionError
from botocore.parsers import ResponseParserError

from .init import CFG
from .retry import retry
from .goodies import b_if_not_a

# ############# INCLUDE HERE THE NAMES OF TESTED MESSAGE QUEUE TYPES ################ #
#   These form the possible choices for:
#          CFG.D["Dependencies"]["message-queue"]["targets"][<host-name>]['type']
#
AWS_SQS_NAME = 'sqs.amazonaws.com'
SupportedMessageQueueTypes = [AWS_SQS_NAME, 'elastic-mq']
# ################################################################################### #


def check_sqs_config(on_err=False):

    target = CFG.D["Dependencies"]["message-queue"]["useTarget"]
    scfg = CFG.D["Dependencies"]["message-queue"]["targets"].get(target) or {}

    if check_sqs_name(target) == AWS_SQS_NAME:
        adv_config = boto3config(user_agent=CFG.user_agent,
                                 connect_timeout=10,
                                 read_timeout=5)
        mq_client = boto3.resource('sqs', config=adv_config)
        scfg.update({"client": mq_client})

    if scfg is None:
        if not on_err:
            print(f"[WARNING] No message-queue config named '{CFG.D['Dependencies']['message-queue']['useTarget']}' was found")
            return None, None
        else:
            raise ValueError(f"No message-queue config named '{CFG.D['Dependencies']['message-queue']['useTarget']}' was found")
    else:
        return scfg, target


def check_sqs_name(name):
    sqslist = ['awssqs', 'sqs', 'sqsaws', 'amazonsqs', AWS_SQS_NAME]
    if isinstance(name, str) and name.lower() in sqslist or any([1 for x in sqslist if x in name.lower()]):
        return AWS_SQS_NAME
    else:
        return name


def _queue_object_from_client(mq_client, queue, op='get'):
    errs = []
    q_obj = None
    try:
        # Depending here on getting a 'QueueDoesNotExist' error.  Generating a random
        # queue name so we can avoid race conditions with other services doing the same check.
        if op == 'create':
            q_obj = mq_client.create_queue(QueueName=queue)
        else:
            q_obj = mq_client.get_queue_by_name(QueueName=queue)
    except EndpointConnectionError as err:
        # Here if the specified url point to a non-existent host
        errs.append(f"Failed to connect to message queue: {err}")
    except (ResponseParserError, ClientError) as err:
        # Elastic-MQ returns the below html when the queue does not exist and the boto3 SQS
        #      client spews "botocore.parsers.ResponseParserError: Unable to parse response"
        #      b'<!doctype html><title>404 Not Found</title><h1 style="text-align: center">
        #      404 Not Found</h1>'
        # AWS SQS throws a botocore Client Error with text "<title>404 Not Found"
        queue_not_found = ["<title>404 Not Found", "AWS.SimpleQueueService.NonExistentQueue"]
        if not any([err.args[0].count(m) for m in queue_not_found]):
            errs.append(f"Unknown error returned from message queue: {err}")
    finally:
        return q_obj, errs


# Returns client only if we can connect to the source_name and the bucket is in the source_name and configured
@retry(RuntimeError, total_tries=5, initial_wait=3, backoff_factor=1.75)
def _message_queue_create_client(on_err=False):
    # Step1: Get the message queue source configuration
    src_config, cfg_name = check_sqs_config(on_err=on_err)
    if src_config is None:
        return None

    host_type = src_config.get('type', '')
    if check_sqs_name(host_type.lower()) not in SupportedMessageQueueTypes:
        print(f"[CONFIG ERROR] In config 'CFG[GLOBAL][csiMessageQueue][{cfg_name}]': "
              f"Unsupported message queue type '{host_type}' ")
        src_config.update({'client': False})
        return None

    mq_client = src_config.get('client', None)

    # Step 2: Set up a new client for the message queue source if it has none
    if mq_client is None or mq_client == 'retry':
        if host_type.lower().startswith('elastic'):
            urls = b_if_not_a(src_config.get('url'), src_config.get('urls'))
            if urls is None:
                print(f"[CONFIG ERROR] No URL configured for Elastic-MQ host '{cfg_name}' ")
                src_config.update({'client': False})
                raise ValueError(f"No URL configured for Elastic-MQ host '{cfg_name}' ")

            if isinstance(urls, str):
                urls = [urls]

            errs = []
            found = False
            for u in urls:
                try:
                    adv_config = boto3config(s3={'addressing_style': 'path'},
                                             user_agent=CFG.user_agent,
                                             user_agent_extra=None,
                                             connect_timeout=10,
                                             read_timeout=20)

                    mq_client = boto3.resource('sqs', endpoint_url=u,
                                               region_name='elasticmq',
                                               use_ssl=u.startswith('https'),
                                               aws_access_key_id=src_config.get('accessKey', ''),
                                               aws_secret_access_key=src_config.get('secretKey', ''),
                                               config=adv_config)
                except Exception as err:
                    errs.append(f"[WARNING] Failed to instantiate client for {host_type} host '{cfg_name}': {err}")
                    src_config.update({'client': False})
                else:
                    q, e = _queue_object_from_client(mq_client, CFG.new_id())
                    if not e:
                        found = True
                        src_config.update({'client': mq_client})
                        src_config.update({'url': u})
                        break

            if not found:
                raise RuntimeError(f"Cannot connect to any Elastic-MQ host in {urls}:\n{errs}")

    return mq_client


notification_bucket_policy = '{"Version": "2012-10-17", "Statement": [{"Sid": "AWSCloudTrailAclCheck20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "s3:GetBucketAcl", "Resource": "%bk_arn%"}, {"Sid": "AWSCloudTrailWrite20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "s3:PutObject", "Resource": "%ct_arn%", "Condition": {"StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}}}, {"Sid": "AWSLogDeliveryWrite", "Effect": "Allow", "Principal": {"Service": "delivery.logs.amazonaws.com"}, "Action": "s3:PutObject", "Resource": "%cw_arn%", "Condition": {"StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}}}, {"Sid": "AWSLogDeliveryAclCheck", "Effect": "Allow", "Principal": {"Service": "delivery.logs.amazonaws.com"}, "Action": "s3:GetBucketAcl", "Resource": "%bk_arn%"}]}'
notification_topic_policy = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__default_statement_ID", "Effect": "Allow", "Principal": {"AWS": "*"}, "Action": ["SNS:GetTopicAttributes", "SNS:SetTopicAttributes", "SNS:AddPermission", "SNS:RemovePermission", "SNS:DeleteTopic", "SNS:Subscribe", "SNS:ListSubscriptionsByTopic", "SNS:Publish", "SNS:Receive"], "Resource": "%topic_arn%", "Condition": {"StringEquals": {"AWS:SourceOwner": "%aws_id%"}}}, {"Sid": "AWSCloudTrailSNSPolicy20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "SNS:Publish", "Resource": "%topic_arn%"}]}'
event_queue_policy_sns = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__owner_statement", "Effect": "Allow"," Principal": {"AWS":"*"},"Action": "SQS:*", "Resource": "%queue_arn%"}, {"Sid": "topic-subscription-%topic_arn%", "Effect": "Allow", "Principal": {"AWS":"*"}, "Action": "SQS:*", "Resource": "%queue_arn%", "Condition": {"ArnLike": {"aws:SourceArn": "%topic_arn%"}}}]}'
event_queue_policy = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__owner_statement", "Effect": "Allow"," Principal": {"AWS":"*"},"Action": "SQS:*", "Resource": "%queue_arn%"}]}'


def get_queue_object_from_name(queue_name):

    if isinstance(queue_name, str) and queue_name:
        # If given a URL, the last element in the path is the queue name,
        # e.g., 'http://csi-dep-queue:9324/queue/csi-s3event-q'
        us = urlsplit(queue_name)
        queue_name = us.path.rsplit('/', 1)[-1]
    elif type(queue_name).__qualname__ == 'sqs.Queue':
        return queue_name  # In this case the argument queue_name is a Queue object, so return it

    queue_client = _message_queue_create_client()
    q_obj, err = _queue_object_from_client(queue_client, queue_name, op='get')

    if not q_obj and not err:
        q_obj, err = _queue_object_from_client(queue_client, queue_name, op='create')

    if err and isinstance(err, list):
        for e in err:
            print(e)

    return q_obj


def delete_queue(queue_name):
    qa = get_queue_object_from_name(queue_name)
    if qa:
        try:
            print(f"Deleting message queue '{qa.url}'")
            ret = qa.delete()
        except ClientError as err:
            ret = {'err': err}
        return ret
    else:
        return


def purge_queue(queue_name):
    qa = get_queue_object_from_name(queue_name)
    if qa:
        try:
            surl = urlsplit(qa.url)
            print(f"Purging message queue '{surl.path.split('/')[-1]}'")
            ret = qa.purge()
        except ClientError as err:
            ret = {'err': err}
        return ret
    else:
        return


def post2queue(queue_name, message=''):
    qa = get_queue_object_from_name(queue_name)
    if qa:
        try:
            sqs_resp = qa.send_message(MessageBody=message)
        except ClientError as err:
            sqs_resp = {'err': err}
        return sqs_resp
    else:
        return

def get_queue(name="APIshdwQueue"):
    q_name = CFG.D["Dependencies"]["message-queue"].get("queues", {}).get(name)
    if q_name is None:
        target = CFG.D["Dependencies"]["message-queue"].get("useTarget")
        q_name = CFG.D["Dependencies"]["message-queue"].get("targets", {})\
                                                       .get(target, {})\
                                                       .get('queues', {})\
                                                       .get(name)
    return  q_name




# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  If used, attribution of open-source code is included where required by original author          #
#                                                                                                  #
#  Filename:  tf_config_load.py                                                                    #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

"""
    Terraform put configurations for each dependent service in environment variables.
    Get what input there is and update the target deependency configuration.

    If the environment variable's value is:
         'DISABLED'      =>   This dependency is not configured to be used so skip the initialization and discovery.

    If the environment variable's value startswith '{' or '[{', then do a json.loads and update the configuration
    for the target with the contents.
"""


import os
from json import loads, JSONDecodeError
from requests import get as req_get
from requests.exceptions import ConnectionError, InvalidSchema


def update_from_terraform(cfg):
    # Becasue of cycles and 'after apply' problems in Terraform, sometimes it's not possible to set
    # CSI_DEPENDENCY_CONFIG.  If that's the case, the Terraform script must write the dependencyConfig
    # contents to a config file in the Caber's main bucket.  After merging configs, CFG in inity.py will
    # call this function to decode the dependency string.
    dep_envars = os.getenv("CSI_DEPENDENCY_CONFIG", cfg.get("GLOBAL", {}).pop("dependencyConfig", None))
    if dep_envars and cfg:
        try:
            dep_envars = loads(dep_envars)
        except JSONDecodeError as err:
            print(f"[WARNING] Error decoding environment variable 'CSI_DEPENDENCY_CONFIG'. Using bootstrap default config.\n{err}")
        else:
            for k, v in dep_envars.items():
                if k.lower() in ['kafka', 'zookeeper']:
                    k = 'kafka-zookeeper'

                tgt_name = v.get("hostname", k)
                try:
                    cfg.D["Dependencies"][k].update({"useTarget": tgt_name})
                    existing = cfg.D["Dependencies"][k]["targets"].get(tgt_name) or {}
                except AttributeError:
                    cfg["Dependencies"][k].update({"useTarget": tgt_name})
                    existing = cfg["Dependencies"][k]["targets"].get(tgt_name) or {}

                for vk in v.keys():
                    existing.update({vk: v[vk]})
                if "ipv4" not in existing.keys():
                    existing.update({"ipv4": None})    # Keep it empty so that it can be updated later in OpenSearch.

                scheme = existing.get('scheme', 'http')
                if "neo4j" in k.lower() and isinstance(scheme, str) and not scheme.startswith("http"):
                    if existing.get('scheme', 'http').count('s'):
                        scheme = "https"
                    else:
                        scheme = "http"

                urls = []
                if isinstance(existing.get("urls"), list) and existing["urls"]:
                    urls = existing.get("urls")
                elif isinstance(existing.get("url"), list) and existing["url"]:
                    urls = existing.get("url")
                elif isinstance(existing.get("url", 0), str) and existing["url"]:
                    urls = [existing["url"]]

                if existing.get('host_full'):
                    if '://' not in existing.get('host_full'):
                        urls.append(f"{scheme}://{existing.get('host_full')}")
                    else:
                        urls.append(existing.get('host_full'))

                # Check if the hostname is that of an AWS EC2 instance.  If so, send a request to it to see if the
                # dependency is running on the same host.  If it responds on the right port
                # add the hostname to the URLs.
                envhost = os.getenv("HOSTNAME", "")
                if envhost.startswith("ip-") and envhost.endswith("compute.internal") and existing.get("port", False):
                    new_url = f"{scheme}://{os.getenv('HOSTNAME')}:{existing.get('port')}"
                    try:
                        req_get(new_url)
                    except InvalidSchema:
                        print(f"[DEBUG] Invalid schema {scheme} for {k}")
                    except ConnectionError:  # Must be from requests.exceptions
                        pass
                    else:
                        urls.append(new_url)

                existing.update({"urls": list(set(urls))})
                try:
                    cfg.D["Dependencies"][k]["targets"].update({tgt_name: existing})
                except AttributeError:
                    cfg["Dependencies"][k]["targets"].update({tgt_name: existing})
    return cfg



# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  goodies.py                                                                           #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

"""
A common location for a number of convenience functions
"""

from urllib.parse import urlsplit
from fnmatch import filter as fnfilter
from itertools import chain
import json
import re
import os
from socket import getfqdn
from collections import Counter
from ipaddress import ip_address, IPv4Address, IPv6Address, AddressValueError

from botocore.utils import is_valid_endpoint_url
from ..Toolbox.aws_init import check_s3_name, get_aws_regions, AWS_S3_NAME, AWS_S3_SHORT_NAME

try:
    from pandas import isna, DataFrame, Series
except ImportError:
    DataFrame = None
    Series = None
    isna = None


def normalize_host_config(scfg: dict, trgt, svc_domain='', auth=False, force=False):
    """
    Normalized one host configuration to a standard form.  Won't normalize if the configuration has already
    been normalized, or if it has a key enabled with value False, unless 'force' is set to True.

    Standard host configuration dictionaries may have the following keys and others:
    keys = {'scheme', 'host', 'port', 'accessKey', 'secretKey', 'use_ssl', 'is_up', 'enabled', 'normalized', 'key'
            'initialized', 'region', 'hostname', 'ipv4', 'ipv6', 'host_full', 'domain', 'serviceName', 'upstrm'}

    scheme is one of 'http', 'https', 'file', a custom scheme like 'bolt' for neo4j, or None

    :param svc_domain: The value of CFG.svc_domain, if available. Default is ''
    :param scfg: The input host configuration dictionary.
    :param trgt: The target host if caller of this function is iterating over a set of {target: {scfg} } dicts.
    :param auth: Should auth keys 'accessKey', 'secretKey', be used to build a URL with basic http auth.
    :param force:
    :return: normalized scfg dictionary.
    """

    norm = scfg.get("normalized", False)
    if not force and norm or not scfg.get("enabled", True):
        return scfg

    scfg = shp_to_port(scfg)
    scfg = _step1_norm_host_key(scfg, trgt, svc_domain=svc_domain)
    scfg = _step2_check_ssl(scfg)
    scfg = _step3_norm_scheme(scfg)
    scfg = _step4_norm_url_urls(scfg)

    if not scfg.get('host') or not scfg.get('host_full'):
        raise ValueError(f"Empty host! {scfg}")
    if ':' in scfg.get('host_full', ''):
        hf = scfg['host_full'].rsplit(':',1)[0]
        scfg.update({'host_full': hf})

    scfg.update({"normalized": True})
    return scfg


def most_common(lst: list):
    """
    Finds the most often occurring value in the input list that is not None or an empty string. or the first
    if no value occurs more than once.
    :param lst: List of values that may have duplicates.
    :return: The most often occurring value in the list
    """

    # Eliminate None and empty string values, then normalize remaining values as strings
    lst = [str(item) for item in lst if item is not None and item != '']
    # If the list is empty after filtering, return None
    if not lst:
        return None
    # Count occurrences of each value in the list
    counts = Counter(lst)
    # Find the most common value (returns a list of tuples (value, count))
    most_common_data = counts.most_common(1)
    # If the most common value has a count of 1, return the first value of the list
    if most_common_data[0][1] == 1:
        return lst[0]
    # Otherwise, return the most common value
    return most_common_data[0][0]


def is_ip_url(url, value_only=True):
    """
    If running on an EC2 instance the hostname can be something like
    'http://ip-10-0-0-220.us-west-1.compute.internal:9200'. If so, extract the Ipv4 address and return.

    :param url:
    :param value_only: return just the ip address if Ture.  Othersise, return dict of {'ipv4': ip}
    :return:
    """

    url = url.split('://')[-1]
    url = url.rsplit(':', 1)[0]
    ip_addr = check_ip_address(url, no_zero=True, value_only=value_only)
    if ip_addr:
        return ip_addr

    pattern = r'^(?:[a-z]{0,10}-?)(\d{1,3}[-|\.]\d{1,3}[-|\.]\d{1,3}[-|\.]\d{1,3})'
    match = re.search(pattern, url)

    if match:
        ip_addr = match.group(1).replace('-', '.')
        return check_ip_address(ip_addr, no_zero=True, value_only=value_only)

    return '' if value_only else {}


# If pandas is available use it to find if a value is a dataframe, series, or empty dataframe
def b_if_not_a(a, b):
    a_is_empty_df = False
    a_is_na = False
    if DataFrame is not None:
        a_is_empty_df = isinstance(a, (DataFrame, Series)) and a.empty
        a_is_na = isna(a)
    a_is_empty_other = not isinstance(a, (bool, int, float)) and not a
    if a_is_empty_df or a_is_na or a_is_empty_other:
        return b
    else:
        return a


def first_valid(*args, blank='evaluates_false', oftype=str):
    """
    Returns the first valid argument passed to the function where valid is determined by
    the parameters 'blank' and 'oftype'.  'blank' determines the method by which an argument A is deemed valid.

    When param 'blank' is:
        'evaluates_false': If the expression 'bool(A)' for arg A returns False, A is not valid.
        'is_none':         If the expression 'A is not None' for arg A returns False, A is not valid.
        None:              Same as 'is_none'
        <list>:            If oftype is not 'list', and if A is not in list 'blank', A is not valid
        <dict>:            If oftype is not 'dict', and if A is not in list 'blank.values()', A is not valid
        <function>:        A is valid if function(A) returns True

        If none of the above applies, then A is considered valid so long as A != 'blank'

    :param args: any number of arguments of any type
    :param blank: Can be one of the command strings 'evaluates_false' or 'is_none', a function that
                  returns True if valid, a list or dict of valid values. Default is 'evaluates_false'

    :param oftype: If set to a type, a in args is considered valid only if isinstance(a, type) is True
    :return: first valid arg or None
    """

    def _is_none(x):
        return x is None

    def _is_not_equal(x):
        if isinstance(x, type(blank)):
            return x != blank
        else:
            return True

    def _is_not_in(x):
        if isinstance(blank, list) and not isinstance(x, list):
            return x not in blank
        elif isinstance(blank, dict) and not isinstance(x, dict):
            return x not in list(blank.values())
        else:
            return _is_not_equal(x)

    if len(args) == 1 and isinstance(args[0], list):
        args = args[0]

    if isinstance(oftype, type):
        args = [a for a in args if isinstance(a, oftype)]

    if blank is None or (isinstance(blank, str) and blank.lower() == 'is_none'):
        out = filter(_is_none, args)
    elif isinstance(blank, str) and blank.lower() == 'evaluates_false':
        out = filter(None, args)
    elif isinstance(blank, type(_is_none)):
        # Here if 'blank' is a function
        # noinspection PyTypeChecker
        out = filter(blank, args)
    elif isinstance(blank, list) or isinstance(blank, dict):
        out = filter(_is_not_in, args)
    else:
        out = filter(_is_not_equal, args)

    out = list(out)
    if out:
        return out[0]
    else:
        return None


def _step2_check_ssl(out: dict) -> dict:
    if out.get("use_ssl", None) is None:
        if out.get("scheme", "") == 'https':
            out.update({"use_ssl": True})
        elif str(out.get("port", "")) == "443":
            out.update({"use_ssl": True})
        else:
            out.update({"use_ssl": False})
    return out


def _step4_norm_url_urls(d: dict) -> dict:
    utemp = []
    utemp.extend(shp_to_url(d, include_auth=False))

    if isinstance(d.get('url'), list):
        utemp.extend(d.pop('url', []))
    elif isinstance(d.get('url'), str):
        utemp.append(d.pop('url'))
    if isinstance(d.get('urls'), list):
        utemp.extend(d.pop('urls', []))
    elif isinstance(d.get('urls'), str):
        utemp.append(d.pop('urls'))

    utemp = [u for u in utemp if isinstance(u, str) and u]
    utemp.sort(key=len)

    mc_url = most_common(utemp)
    if mc_url:
        goodurls = [] if is_ip_url(mc_url) else [mc_url] if mc_url else []
        # Order such that longest non-ip URL is first.
        goodurls.extend(list(set([u for u in utemp if isinstance(u, str) and u and not is_ip_url(u) and u != mc_url])))
        goodurls.sort(key=len, reverse=True)
        goodurls.extend(list(set([u for u in utemp if isinstance(u, str) and u and is_ip_url(u)])))

        goodurls = [
            u for u in goodurls
            if not (
                u.endswith(os.getenv("CSI_ROLE_EXTERNAL_ID", "")) or
                u.endswith(os.getenv("CSI_TENANT_ID", "")) or
                u.endswith(os.getenv("CSI_DEPLOYMENT_ID", ""))
            )
        ]

        d.update({'urls': goodurls})

    else:
        d.update({'urls': []})

    return d


def _step3_norm_scheme(d: dict) -> dict:
    default_scheme = 'https' if str(d.get('port', 0)) == "443" or d.get('use_ssl', False) else 'http'
    scheme = first_valid(d.get('scheme'), d.get('protoForXFProto'), 
                         most_common([urlsplit(p).scheme for p in d.get('urls', [])
                                      if p and isinstance(d.get('urls'), list)]), default_scheme)
    if scheme:
        d.update({"scheme": scheme})
    return d


def _step1_norm_host_key(d: dict, key: str = '', svc_domain='') -> dict:
    """
    Every configured host has a standard dictionary of values. From this dictionary determine if
    the 'host' is an AWS service.  If so, then make sure the host and host_full keys in the host
    dictionary are standardized so their use will be consistent everywhere.

    If not an AWS service, the host configuration parameters can be specified in a number of ways.
    Through many iterations I found forcing a one-size-fits-all config for hosts was nearly
    impossible to maintain as new hosts types were added. So, instead this function allows flexibility
    by deconstructing every value related to the name of the host from the host config dictionary into
    schemes, prefixes, hostnames, instance numbers, domains, and ports.  Then it reconstructs the
    needed parameters from this base set.

    Priority is given to values explicitely stated.  For example if the key {"scheme": "http"}
    exists then it's prioritized over the scheme 'https' in {'urls': ['https://hostname']}.

    NOTE: On docker networks, a DNS lookup of the container name yields a valid result, but so does
    prefix-containerName-instance.  So do a lookup for the fully-qulaified domain name using the
    shortest valid values for containerName to find the FQDN.  If we have the FQDN and the container
    name then extract the prefix, instance, domain, etc.  Check against values specified in the
    host config dictionary.

    :param d:
    :return:
    """

    keys = {'scheme', 'host', 'port', 'accessKey', 'secretKey', 'use_ssl', 'is_up', 'url',
            'urls', 'enabled', 'normalized', 'key', 'initialized', 'region', 'hostname',
            'ipv4', 'ipv6', 'host_full', 'domain', 'serviceName', 'upstrm', 'prefix', 'path'}
    keeps = keys.intersection(d.keys())
    domain = ''
    elba = d.get('load_balancer_arn', '')

    key2 = key if 'self' not in key else ''
    raw = [d.get('host'), key2, d.pop('key', ''), d.get('upstrm'), d.pop('host_full', ''),
           d.pop('hostname', ''), d.get('serviceName'), d.get('url')]
    raw2 = [r for r in raw if r and isinstance(r, str) and is_ip_url(r)]
    raw1 = [r for r in raw if r and isinstance(r, str) and not is_ip_url(r)]
    raw = [url_to_shp(r, prt=d.get('port', '')) for r in raw1]

    if any([1 for r in raw if check_s3_name(r.get('hostname', '')) == AWS_S3_NAME]):
        host = AWS_S3_SHORT_NAME
        hfull = AWS_S3_NAME
        d.update({"host": host})
        d.update({"host_full": hfull})
        d.update({"domain": "amazonaws.com"})
    elif elba:
        host = d.get('serviceName') or first_valid([r.get('hostname', '') for r in raw])
        hfull = d.get('upstrm') or first_valid([r.get('hostname', '') for r in raw])
        d.update({'public_dns_name': first_valid(key2, *[r.get('hostname', '') for r in raw])})
        port = first_valid([str(r.get('port', '')) for r in raw])
        d.pop('scheme', '')
        d.pop('domain', '')

        raw = [d.get('upstrm'), d.pop('host_full', '')]
        raw1 = [r for r in raw if r and isinstance(r, str) and not is_ip_url(r)]
        raw = [url_to_shp(r, prt=d.get('port', '')) for r in raw1]
        scheme = first_valid([r.get('scheme', '') for r in raw])
        domain = first_valid(*[r.get('domain', '').strip('. -') for r in raw],svc_domain.strip('. -'))
        hfull = url_to_shp(hfull).get('hostname', hfull)
        d.update({"host": host.strip('. -')})
        d.update({"host_full": hfull.strip('. -')})
        d.update({"port": port})
        d.update({"scheme": scheme})
        d.update({'domain': domain})

    else:
        host = first_valid([r.get('hostname', '') for r in raw])
        port = first_valid([str(r.get('port', '')) for r in raw])
        scheme = first_valid(d.get("scheme", ''), *[r.get('scheme', '') for r in raw])
        domain = first_valid(d.get("domain", '').strip('. -'), *[r.get('domain', '').strip('. -') for r in raw],
                             svc_domain.strip('. -'))

        d.update({"port": port} if port else {})
        d.update({"scheme": scheme} if scheme else {})
        d.update({"domain": domain} if domain else {})
        prefix = d.get('prefix', '') or ''

        if host:
            fqdn = getfqdn(host)
            if len(fqdn) > len(host) and host in fqdn:
                dpp = fqdn.split(host, 1)
                # Check if there is an instance number
                if dpp[-1].startswith('-') and '.' in dpp[-1]:
                    dsplt = dpp[-1].strip('-').split('.', 1)
                    # Add the instance number if it exists
                    d['instance'] = dsplt[0] if dsplt[0].isdecimal() else d.get('instance') or ''
                    dpp[-1] = dsplt[1] if dsplt[0].isdecimal() else dpp[-1].strip('-')

                elif domain and dpp[-1].strip('. -') != domain:
                    print(f"[DEBUG] getfqdn returned domain {dpp[-1].strip('. -')} but domain {domain} was specified")

                domain = dpp[-1].strip('. -')
                d.update({"domain": domain} if domain else {})

                if len(dpp) > 1 and prefix and dpp[0].strip('. -') != prefix.strip('. -'):
                    print(f"[DEBUG] getfqdn returned prefix {dpp[-1].strip('. -')} but prefix {prefix} was specified")
                elif len(dpp) > 1:
                    prefix = dpp[0].strip('. -')
                    d.update({"prefix": prefix} if prefix else {})
        else:
            host = first_valid([url_to_shp(r).get('hostname', '') for r in raw2])

        port = d.get('port', '') or ''
        domain = d.get('domain', '') or ''
        prefix = d.get('prefix', '') or ''
        instance = d.get('instance', '') or ''

        host = host.split(':', 1)[0] if ':' in host else host
        host = host.replace(domain, '').strip('. -') if domain and domain in host else host
        host = host.replace(prefix, '').strip('. -') if prefix and prefix in host else host
        host = host.replace(instance, '').strip('. -') if instance and instance in host else host

        hfull = '-'.join([prefix, host, instance]).strip('. -')
        hfull = '.'.join([hfull, domain]).strip('.')
        if port and port != '443' and port != '80':
            hfull = ':'.join([hfull, port]).strip(':')

        if "amazonaws.com" in host or "amazonaws.com" in hfull:
            d.update({"domain": "amazonaws.com"})

        if "amazonaws.com" in host:
            d.update({"host": host.replace('.amazonaws.com', '').strip('. -')})

        if isinstance(hfull, str) and hfull:
            d.update({"host_full": hfull.strip('. -')})

        if isinstance(host, str) and host:
            if (host == os.getenv('CSI_ROLE_EXTERNAL_ID', '') or
                    host == os.getenv('CSI_TENANT_ID', '')  or
                    host == os.getenv('CSI_DEPLOYMENT_ID', '') or
                    host.count("newcopet") or hfull.count("newcopet")):
                print(f"[DEBUG] host {host} is a CSI ID.  Setting host to newco-nextcloud.caber.svc")
                d.update({"host": "newco-nextcloud.caber.svc"})  #FIXME: HACK!
                d.update({"hostname": None})
            else:
                d.update({"host": host.strip('. -')})

    return d


def url_to_shp(url: str, prt='', host='') -> dict:
    """
    Convert a url to a dict with scheme, hostname, and port.  Used mostly for validating configuration params.
    Discards any path or query.  After a reverse DNS lookup the url can be something like
    'http://ip-10-0-0-220.us-west-1.compute.internal:9200'. If so, extract the Ipv4 address, domain and return.

    This function purposely does not set keys 'host' or 'host_full' to avoid the calling function from
    potentially overwriting existing values for these keys.  Instead, it writes 'hostname' and leaves it to the
    calling function to replace host or host_full if needed.

    :param url: string representation of a url
    :return: dict with ['scheme', 'hostname', 'port', 'accessKey', 'secretKey', 'domain', 'region', 'ipv4', 'ipv6']
             keys as appropriate for the given input.
    """
    ret = {}
    domain = ''

    if isinstance(url, (str, bytes, bytearray)) and url:
        if '://' not in url:
            spl_url = urlsplit('foo://' + url)
        else:
            spl_url = urlsplit(url)

        pattern = r'^(?:[a-z]{0,10}-?)(\d{1,3}[-|\.]\d{1,3}[-|\.]\d{1,3}[-|\.]\d{1,3})'
        match = re.search(pattern, spl_url.hostname)

        if match:
            ip_addr = match.group(1).replace('-', '.')
            domain = spl_url.hostname.replace(match.group(0), '').strip('. -')
            ret = check_ip_address(ip_addr, no_zero=True, value_only=False)
            ret.update({"hostname": match.group(0)})
        else:
            ret = check_ip_address(spl_url.hostname, no_zero=True, value_only=False)
            if not ret and spl_url.hostname:
                ret = {"hostname": spl_url.hostname.strip('. -')}

        for region in get_aws_regions():
            if domain and region in domain:
                ret.update({"region": region})
                domain = domain.replace(region, '').strip('. -')
                break
            elif not domain and region in spl_url.hostname:
                ret.update({"region": region})
                break

        if (isinstance(prt, str) and prt.isdecimal() or isinstance(prt, int)) and 0 < int(prt) < 65536:
            ret.update({"port": int(prt)})

        elif spl_url.port and str(spl_url.port).isdecimal() and 0 < int(spl_url.port) < 65536:
            ret.update({"port": int(spl_url.port)})

        ret.update({"scheme": spl_url.scheme} if spl_url.scheme != 'foo' else {})
        ret.update({'accessKey': spl_url.username})
        ret.update({'secretKey': spl_url.password})
        ret.update({"domain": domain})

    out =  {k: v for k, v in ret.items() if isinstance(v, bool) or v}

    return out


def check_ip_address(is_ip, no_zero=False, value_only=False):
    """
    Checks for a valid IPv4 or IPv6 address and returns it.
    :param is_ip: The value to test
    :param no_zero: Don't return '0.0.0.0' as a valid IPv4 address
    :param value_only: return just the IP address if True, otherwise return {'ipv4': IP}
    :return:
    """

    try:
        ip = ip_address(is_ip)
    except (ValueError, AddressValueError):
        return '' if value_only else {}
    else:
        ipc = str(ip.compressed).strip()
        if isinstance(ip, IPv4Address):
            if no_zero and ipc == "0.0.0.0":
                return '' if value_only else {}
            return  ipc if value_only else {"ipv4": ipc}
        else:
            return ipc if value_only else {"ipv6": ipc}


def combine_shps(shpa: dict, shpb: dict) -> dict:
    """
    Given two dicts with scheme, hostname, port, etc., that may each be missing some values, create one
    dict with a complete set of values giving precedence to shpa -- that is, if shpa and shpb both have a 
    value for, say, 'host', then shpa[host] will be in the output.  Ensure that the username and password
    are passed as a unit.
    
    :param shpa: dict with {'scheme': '', 'host': '', 'port': '', 'accessKey': '', 'secretKey': ''}
    :param shpb: dict with {'scheme': '', 'host': '', 'port': '', 'accessKey': '', 'secretKey': ''}
    :return: dict with {'scheme': '', 'host': '', 'port': '', 'accessKey': '', 'secretKey': ''}
    """
    
    out = {}

    keys = set(shpa.keys()).union(shpb.keys()).difference(["accessKey", "secretKey", "ipv4", "ipv6"])

    if shpa.get("accessKey", "") and shpa.get("secretKey", ""):
        out.update({'accessKey': shpa["accessKey"], 'secretKey': shpa["secretKey"]})
    elif shpb.get("accessKey", "") and shpb.get("secretKey", ""):
        out.update({'accessKey': shpb["accessKey"], 'secretKey': shpb["secretKey"]})
    elif shpa.get("accessKey", "") and not shpb.get("accessKey", "") or \
            shpa.get("secretKey", "") and not shpb.get("secretKey", ""):
        out.update({'accessKey': shpa.get("accessKey", ""), 'secretKey': shpa.get("secretKey", "")})
    else:
        out.update({'accessKey': shpb.get("accessKey", ""), 'secretKey': shpb.get("secretKey", "")})

    ipv4s = shpa.get("ipv4", []) or []
    ipv4s.extend(shpb.get("ipv4", []) or [])
    ipv4s.append([shpa.get("host") or "", shpb.get("host") or ""])
    ipv4s = [ ip for ip in ipv4s if check_ip_address(ip, no_zero=True, value_only=True)]
    ipv4s = [i for i in ipv4s if i]
    ipv6s = shpa.get("ipv6", []) or []
    ipv6s.extend(shpb.get("ipv6", []) or [])
    ipv6s = [ ip for ip in ipv6s if check_ip_address(ip, no_zero=True, value_only=True)]
    ipv6s = [i for i in ipv6s if i]
    out.update({"ipv4": ipv4s}) if ipv4s else None
    out.update({"ipv6": ipv6s}) if ipv6s else None

    for k in keys:
        result = b_if_not_a(shpa.get(k), shpb.get(k))
        out.update({k: result})
    #
    # out.update({'region': b_if_not_a(shpa.get("region", ""), shpb.get("region", ""))})
    # out.update({'scheme': b_if_not_a(shpa.get("scheme", ""), shpb.get("scheme", ""))})
    # out.update({'host': b_if_not_a(shpa.get("host", ""), shpb.get("host", ""))})
    # out.update({'host_full': b_if_not_a(shpa.get('host_full', ""), shpb.get('host_full', ""))})
    # out.update({'port': b_if_not_a(shpa.get("port", ""), shpb.get("port", ""))})
    # out.update({'open_port': b_if_not_a(shpa.get("open_port", ""), shpb.get("open_port", ""))})
    # out.update({'use_ssl': b_if_not_a(shpa.get("use_ssl"), shpb.get("use_ssl"))})
    # out.update({'is_up': b_if_not_a(shpa.get("is_up"), shpb.get("is_up"))})
    # out.update({'upstrm': b_if_not_a(shpa.get("is_up"), shpb.get("is_up"))})
    # out.update({'initialized': b_if_not_a(shpa.get("initialized"), shpb.get("initialized"))})
    # out.update({'enabled': shpa.get("enabled", False) or shpb.get("enabled", False)})

    return {k: v for k, v in out.items() if v is not None and v != ''}


def shp_to_port(shp: dict):
    """
    Look for a service port in the host configuration (shp).  If not implied by the scheme (that is, 80 or 443),
    the service port should be an integer value under the key 'port'.  However, it may also be specified as the
    first entry in a list of ports under the key 'ports', or, if the key 'ports' contains a dict, then under the
    key 'service'.  If all that fails, see if host_full has a port, or take the first port in a dict that is not
    a debug port, then use it.

    :param shp:  Dictionary containing host configuration parameters.
    :return:  Dictionary containing host configuration parameters with the port key updated.
    """

    def _gsp(x):
        if x:
            a = str(x) if isinstance(x, (str, int)) and str(x).isdecimal() else ''
            b = str(x[0]) if isinstance(x, list) and len(x) > 0 and str(x[0]).isdecimal() else ''
            c = ''
            if isinstance(x, dict) and len(x.keys()):
                vees = [str(v) for k, v in x.items() if k.lower() not in ['debug', 'service'] and str(v).isdecimal()]
                if str(x.get("service", "")).isdecimal():
                    c = str(x["service"])
                elif len(vees):
                    c = vees[0]
            return first_valid(a, b, c)
        return

    p1 = _gsp(shp.get("port"))
    p2 = _gsp(shp.get("ports"))
    p3 = _gsp(shp.get("host_full", "").rsplit(':', 1)[-1])

    service_port = first_valid(p1, p2, p3)
    if service_port:
        shp.update({"port": service_port})
    return shp


def shp_to_url(shp: dict, include_auth=True, rtn='all') -> str:
    """
    Convert a dict with scheme, hostname, and port to a url.  Used mostly for validating configuration params.
    Discards any path or query.  For consistency, always give precedence to the explicit value, i.e., port = 9975, rather than
    the implicit value i.e., hostname:3476.

    :param include_auth: If True and if username or password exist, add them to the url (default=True)
    :param shp: dict with {'scheme': '', 'host': '', 'port': '', 'accessKey': '', 'secretKey': ''}
    :param rtn: If rtn == 'all' the return value will be a list of all valid urls, else it will return only the
                first valid url.
    :return:  string representation of a url
    """
    
    def _construct_url(_hostname, _shp, _include_auth, _scheme):

        hs = _hostname.rsplit(':', 1)
        port = ''
        if len(hs) == 2:
            port = hs[-1]
            host_no_port = hs[0]
        else:
            host_no_port = _hostname

        # Don't add ports 80 or 443 to urls that imply them.  Also keeps us from putting port 80 on an https url.
        proto = b_if_not_a(_shp.get('scheme', ''), _scheme)
        port = '' if str(port) in ["80", "443"] and proto in ["http", "https"] else port
        sport = '' if str(_shp.get('port')) in ["80", "443"] and proto in ["http", "https"] else _shp.get('port', '')
        port = b_if_not_a(sport, port)

        if _include_auth and (_shp.get("accessKey", "") or _shp.get("secretKey", "")):
            url = f"{proto}://{_shp.get('accessKey', '')}:{_shp.get('secretKey', '')}@{host_no_port}"
        else:
            url = f"{proto}://{host_no_port}"

        if port:
            url = f'{url}:{port}'

        if is_valid_endpoint_url(url):
            return url
        return

    # ===================================================================================================

    scheme = "http" if shp.get("use_ssl", False) else 'http'
    scheme = shp.get("scheme", scheme)

    if not shp.get('domain', '') and shp.get('host_full', '').count('.'):
        dtry = shp['host_full'].split('.', 1)[-1]
    else:
        dtry = shp.get('domain', '')

    invalid = "{_}"
    hostname = [shp.get("host_full", invalid),
                shp.get("host_full", "").replace(dtry, '').replace(".:", ":"),
                f"{shp.get('host', invalid)}.{shp.get(dtry, '')}".strip('.: -'),
                shp.get("host", invalid),
                shp.get("ipv4", invalid),
                shp.get("ipv6", invalid)]

    ids_to_drop = [os.getenv("CSI_ROLE_EXTERNAL_ID"), os.getenv("CSI_TENANT_ID"), os.getenv("CSI_DEPLOYMENT_ID")]
    hostname = [h for h in hostname if h not in ids_to_drop]

    urls = [_construct_url(h, shp, include_auth, scheme) for h in hostname if h and isinstance(h, str)]

    if rtn != 'all':
        urls = first_valid(urls)
    else:
        urls = [u.strip('. -') for u in urls if u]
        urls = list(set(urls))

    if not urls:
        print(f"[WARNING] shp_to_url: No valid hostname found to construct url")
        return "" if rtn != 'all' else []

    return urls


def fill_shp_from_url(url: str, shp: dict, reverse=False) -> dict:
    """
    Explode url in standard host dictionary and fill in the standard
    :param url:
    :param shp:
    :param reverse:
    :return:
    """
    port = shp.get('port', '')
    if reverse:
        return combine_shps(shp, url_to_shp(url, port))
    else:
        return combine_shps(url_to_shp(url, port), shp)


def fnfilter_list(input_list: list, pattern_list: list):
    ret = chain.from_iterable(fnfilter(input_list, pat) for pat in pattern_list)
    return list(ret)


def split_string_list(ss: str, delim=','):
    sl = [ss]
    if ss.startswith('[') and ss.endswith(']'):
        try:
            sl = json.loads(ss)
        except json.JSONDecodeError as err:
            print(f"[WARNING] Ignoring Bad JSON list: {err}")

    elif ' ' in ss or delim in ss:
        ss = ss.replace(' ', delim)
        sl = [s.strip(delim + '[]{}()') for s in ss.split(delim) if s]

    return sl




# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  remote_open.py                                                                       #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################


"""
This module attempts to abstract away the complexities of dealing with multiple types of AWS_S3-like object
store services.  Even though https://pypi.org/project/smart-open/ gives us the ability to read and write objects
and files from many types of sources, there are source-dependent quirks that have to be dealt with.

For example, although Min.io is supposed to be completely S3 compatible, it requires a username:password
combo to be passed and the S3 client must have the URL of the Minio host defined.

The other quirk is that for use internal to Caber services we can use AWS_S3 or another source, and customers
can define multiple types of hosts simultaneously.  All of these we keep track of here.

First thing we need to do is patch smart-open's parse_uri function so that it recognizes our schemes that
look like

    pointer<CFG.token>bucket/prefix    where pointer=pointer to Object_Source config

instead of

    scheme://bucket/prefix

"""

from time import sleep
from urllib.parse import urlsplit
from pathlib import Path, PurePath, PosixPath
from fnmatch import fnmatch
from smart_open import open as smart_open_open

from botocore.exceptions import ClientError, EndpointConnectionError
from botocore.parsers import ResponseParserError
from urllib3.exceptions import NewConnectionError

from .init import CFG, post_status_message
from .retry import retry
from .goodies import first_valid, normalize_host_config, fnfilter_list
from .aws_init import boto3, boto3config, AWS, check_s3_name, AWS_S3_NAME


# ############# INCLUDE HERE THE NAMES OF TESTED OBJECT STORE TYPES ################ #
#   These form the possible choices for:
#          CFG.D['Object_Sources'][<host-name>]['type']
#          CFG.D["Dependencies"]["shared-storage"]["targets"][<host-name>]['type']
#
# AWS_S3_NAME = 's3.amazonaws.com'
SupportedS3likeStoreTypes = [AWS_S3_NAME, 'minio']
cs3idp = f"cs3aws"
# ################################################################################## #


default_configs = {AWS_S3_NAME: {"type": "s3",
                                 "url": "https://s3.amazonaws.com",
                                 "smartOpenPfx": "s3://",
                                 "client_type": "boto3.s3",
                                 "client": None,
                                 "transport_params": None
                                 },
                        "minio": {"type": "minio",
                                  "url": "http://csi-dep-minio:9000",
                                  "smartOpenPfx": "s3://",
                                  "region": "",
                                  "accessKey": "<user>",
                                  "secretKey": "<psswd>",
                                  "client_type": "boto3.s3",
                                  "client": None,
                                  "transport_params": {"client": None}
                                  }
                   }


# TODO: Implement object versions, Implement decryption for encrypted objects


def _build_map(host='*', port='*', bkt_lst=None, id_prefix=''):
    if not id_prefix:
        id_prefix = CFG.D["Dependencies"]["shared-storage"].get("prefix", "gss")
    if not bkt_lst:
        bkt_lst = ['*']
    elif isinstance(bkt_lst, str):
        bkt_lst = [bkt_lst]

    if ':' in host:
        hs = host.split(':')
        host = hs[0]
        if hs[1]:
            port = hs[1]

    # Make an entry in the host map for each bucket with hostname with and without the port
    # The keys in the map will be fnmatch patterns so the lookup can take a wider variety of
    # imputs.  For example: looking up 'http://myhost:80/mybucket/someobject/somewhere'
    # when the hostMap key is '*myhost*mybucket*' will match.

    # If a port is specified, then we are saying port must be used to match.  This would only
    # be required in the case of two storage sources on the same host but with different ports.
    # Here we set up a exact match list to see if we've put the key in before.  We enter patterns
    # (emap) potentially with wildcards in the check_overlaps list, but compare them against
    # values (chke) that have wildcards replaces by ' '.
    emap = [f"{host}:{port}:{c}:{id_prefix}" for c in bkt_lst]
    chke = [f"{host}:{port}:{c}:*" for c in bkt_lst]

    if CFG.D['Object_Sources'].get('check_overlaps'):
        check = fnfilter_list(chke, CFG.D['Object_Sources']['check_overlaps'])
        if check and [c for c in check if c.split(':')[-1] != id_prefix]:
            raise ValueError("We've already inserted a storage host item that matches")
        else:
            CFG.D['Object_Sources']['check_overlaps'].extend(emap)
    else:
        CFG.D['Object_Sources'].update({'check_overlaps': emap})

    hmap = [f"*{host}*{port}*{c}*" for c in bkt_lst]

    if CFG.D['Object_Sources'].get('hostMap'):
        CFG.D['Object_Sources']['hostMap'].update({m: id_prefix for m in hmap})
    else:
        CFG.D["Object_Sources"].update({"hostMap": {m: id_prefix for m in hmap}})

    return


def _lower_key_and_get_config(tcfg: dict, trgt: str):
    if check_s3_name(trgt) == AWS_S3_NAME:
        scfg = tcfg.pop(trgt)
        tcfg.update({AWS_S3_NAME: scfg})
        trgt = AWS_S3_NAME
    elif trgt != trgt.lower():
        scfg = tcfg.pop(trgt)
        tcfg.update({trgt.lower(): scfg})
        trgt = trgt.lower()

    if isinstance(tcfg, dict) and isinstance(tcfg.get(trgt), dict):
        scfg = normalize_host_config(tcfg[trgt], trgt, svc_domain=CFG.svc_domain)
    else:
        print(f"[ERROR] For target {trgt} tcfg[target] is not of type dict")
        scfg = {}
    return scfg, trgt


def _extract_bucket_list(bkts, addin=None) -> list:
    out = []
    if isinstance(bkts, list):
        for b in bkts:
            if isinstance(b, str):
                out.append(b)
            elif isinstance(b, dict):
                out.extend(list(b.keys()))
    elif isinstance(bkts, dict):
        out.extend(list(bkts.keys()))
    elif isinstance(bkts, str):
        out = [bkts]
    if isinstance(addin, str):
        out.append(addin)
    elif isinstance(addin, list):
        out.extend(addin)
    out = set(out)
    out.discard('')
    out.discard(None)
    return list(out)


def _put_bkt_versioning(sclnt, bkt, minio=False):
    '''
    Turn object versioning on in customer bucket, deal with error, retry if necessary.
    TODO: this function should be performed by Terraform in the future so that state is properly maintained.
    :param sclnt: The boto3.client("s3")
    :param bkt: The name of the bucket to put versioning on
    :return: True if operation succeeds.
    '''
    n = 10
    while n > 0:
        n -= 1
        try:
            if minio:
                response = sclnt.put_bucket_versioning(Bucket=bkt,
                                VersioningConfiguration={'MFADelete': 'Disabled', 'Status': 'Enabled'})
            else:  # AWS S3
                response = sclnt.put_bucket_versioning(Bucket=bkt, ChecksumAlgorithm='SHA256',
                                VersioningConfiguration={'MFADelete': 'Disabled', 'Status': 'Enabled'})
        except ClientError as err:
            e = repr(err).split("botocore.exceptions.ClientError: An error occurred (", 1)[-1]
            e = e[:e.find(")")]
            print(f"[WARNING] Boto3 ClientError ({e}) while tyring to set versioning on bucket '{bkt}'")
            if n > 0:
                print(f"[DEBUG] Retrying in 5 sec")
                sleep(5)
        else:
            print(f"[INFO] Enabled versioning on customer bucket {bkt}")
            return True
    return False


def validate_all_store_configs(which='all', force=False):
    """
    Validates the configuration for remote storage in the raw config file and test that we can read and, in
    the case of Caber global shared storage, write to the storage host. Called whenever remote_open is imported.

    :param which: which storage configuration to validate. Default is all
    :param force: Force revalidation of configs if True
    :return: None
    """

    rev_map = {cs3idp: AWS_S3_NAME}
    new_sources = {}

    if not CFG.D.get("Object_Sources", {}):
        CFG.D.update({"Object_Sources": {"revMap": rev_map}})
    elif not CFG.D.get("Object_Sources", {}).get("revMap", ''):
        CFG.D["Object_Sources"].update({"revMap": rev_map})
    else:
        CFG.D["Object_Sources"]["revMap"].update(rev_map)

    # The id_prefix for Caber shared store is always the same: 'gss' for global shared store
    # Note that this default cannot start with 'c' as customer stores will start with that.

    # RANT: Using '://' as the separator between the 'scheme' and 'host' as with urls creates problems
    # when appending the prefix to other strings.  It's hard to ensure there are the proper number of /'s
    # without stripping and re-adding the slashes, and special case-ing file:///directory that always must
    # have three slashes not two. So, internally we will always use the unicode Caber-token (CFG.token) to
    # represent the '://'

    # ---------- START Validating global storage source
    if which == 'all' or which.startswith('g'):

        target = CFG.D["Dependencies"]["shared-storage"]["useTarget"]
        scfg, target = _lower_key_and_get_config(CFG.D["Dependencies"]["shared-storage"].get("targets") or {}, target)
        CFG.D["Dependencies"]["shared-storage"].update({"useTarget": target})

        if not target:
            return False

        print(f"{target.upper()}: Validating configuration (Caber shared storage)")

        # There must be a shared storage bucket name configured
        buckets = CFG.D["Dependencies"]["shared-storage"].get("buckets", '') or ''
        csib = first_valid(CFG.bucket, scfg.get("csiBucket"), CFG.G.get("csiBucket"))
        scfg.update({"csiBucket": csib})
        buckets = _extract_bucket_list(buckets, csib)
        scfg.update({"buckets": buckets})

        idp = CFG.D["Dependencies"]["shared-storage"].get("prefix", "gss")
        scfg.update({"id_prefix": f"{idp}{CFG.token}"})
        # Todo: Support different client types in the future.
        scfg.update({"client_type": f"boto3.s3"})

        _build_map(scfg.get('host', ''), '*', buckets, idp)
        # _build_map(scfg.get('host', ''), scfg.get('port', ''), buckets, idp)

        # Update the storage source config and instantiate the client. Update of revMap must happen
        # before the call to get object store client.
        CFG.D["Object_Sources"]["revMap"].update({idp: target})
        get_object_store_client()

    # ---------- Done Validating global storage source

    # ---------- START Validating customer storage sources
    if which == 'all' or which.startswith('c'):

        oss = [s for s in CFG.D["Object_Sources"].keys() if s not in ["revMap", "doc", "Buckets", 'hostMap', 'check_overlaps']]
        all_customer_buckets = CFG.D["Object_Sources"].get("Buckets", [])

        # below is to check in CFG.D['ApiMap'] to see if any of the sources are being proxied by Caber
        # and, if so, create a configuration for them.
        tap_name = CFG.G["caberModuleContainerNames"]["API_Tap"]
        if isinstance(tap_name, dict):
            tap_name = tap_name.get("host", "")

        tap_name = tap_name.lower()
        taps = {f"{tap_name}:{urlsplit(k).port}": urlsplit(v.get("upstrm", "")).hostname
                for k, v in CFG.D["ApiMap"].items()
                if k.lower().count("self") or k.lower().count(tap_name)}

        new_sources = {}
        for src in oss:
            print(f"Validating customer object source configuration '{src}'")
            scfg, src = _lower_key_and_get_config(CFG.D["Object_Sources"], src)

            buckets = _extract_bucket_list(scfg.get('buckets'))
            all_customer_buckets.extend(buckets)

            if src == AWS_S3_NAME:
                idp = cs3idp
                # new_sources.update({AWS_S3_NAME: default_configs.get(AWS_S3_NAME, {})})

            else:
                idp = src.lower()
                for ch in "/!@#$%^&*()}{_-=+[]<>.,?~`\\|:;'\"":
                    idp = idp.replace(ch, '')
                idp = f"c{idp}"
                if len(idp) > 10:
                    idp = idp[:10]

            scfg.update({"id_prefix": f"{idp}{CFG.token}"})
            scfg.update({"client_type": "boto3.s3"})

            _build_map(scfg.get('host', ''), '*', buckets, idp)

            # Now add any api-tap ports to the list if we happen to be proxying the API to the storage
            # system as we are doing in the demo.
            if scfg.get('host', ''):
                tap_match = [k for k, v in taps.items() if v and v in scfg.get('host', '')]
                for tm in tap_match:
                    _build_map(tm, '', buckets, idp)

            # Update the storage source config and instantiate the client. Update of revMap must happen
            # before the call to get object store client.
            CFG.D["Object_Sources"]['revMap'].update({idp: src})
            sclnt = get_object_store_client(idp)

            # We need bucket versioning to track  data in objects that may be written to multiple times before
            # we get notified of the first write to them.  Here we need to check if bucket versioning is enabled
            # on each bucket and, if it is not, enable it if configuation allows.
            # Minio does not support versioning of objects in 'FileSystem' mode -- one Minio instance and
            # one local disk.  But there is no way to find out if it's in FileSystem mode except try and
            # get the exception.
            envers = scfg.get("enableBktVersioning", False)
            for bkt in buckets:
                try:
                    vers = sclnt.get_bucket_versioning(Bucket=bkt)
                except ClientError as err:
                    print(f"[WARNING] Checking versioning on bucket '{bkt}':  Bucket does not exist")
                else:
                    if envers and (not vers or isinstance(vers, dict) and vers.get('Status', '') != 'Enabled'):
                        response = _put_bkt_versioning(sclnt, bkt, minio=(src == 'minio'))

        CFG.D["Object_Sources"].update({"Buckets": list(set(all_customer_buckets))})

    # ---------- DONE Validating customer storage sources

    return True


class caber_patch_pathlib(object):
    """
    Replace `Path.open` with `remote_open.open`
    Directly from https://github.com/RaRe-Technologies/smart_open/blob/develop/smart_open/smart_open_lib.py
    """

    def __init__(self):
        self.old_impl = _caber_patch_pathlib(ropen)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _caber_patch_pathlib(self.old_impl)


def _caber_patch_pathlib(func):
    """Replace `Path.open` with `func`"""
    old_impl = Path.open
    Path.open = func
    return old_impl


def ropen(*args, **kwargs):

    """
    uri,                                    positional arg
    mode = 'r',                             positional or keyword arg
    buffering = -1,                         positional or keyword arg
    encoding = None,                        positional or keyword arg
    errors = None,                          positional or keyword arg
    newline = None,                         positional or keyword arg
    closefd = True,                         positional or keyword arg
    opener = None,                          positional or keyword arg
    compression = (infer),                  keyword arg
    transport_params = None                 keyword arg
    """

    transport_params = None

    if len(args):
        args = list(args)
        args[0], transport_params = caber_parse_uri(args[0])

    if transport_params is None:
        transport_params = kwargs.get("transport_params", None)
    if transport_params is not None:
        kwargs.update({"transport_params": transport_params})

    try:
        ret = smart_open_open(*args, **kwargs)
    except ValueError as err:
        if str(err).count('ExpiredToken'):
            print(f"[ERROR] Got ValueError->ExpiredToken Exiting")
            post_status_message(stage="4.0", category="warning",
                                message=f"Expired auth token. Shutting down",
                                status="red")
            exit(0)

        print(f"[WARNING] Got ValueError trying to open '{args[0]}': {err}")
        print(f"bucket = {CFG.bucket}")
        return
    except ClientError as err:
        print(f"[WARNING] Got Boto3 ClientError trying to open '{args[0]}': {err}")
        return
    # except (FileExistsError, FileNotFoundError, OSError) as err:
    #     print(f"[WARNING] Got OSError trying to open '{args[0]}': {err}")
    #     return
    else:
        return ret


def caber_parse_uri(uri_as_string):
    """
    Parse a Caber object remote identifier (CORI) and return the corresponding smart-open URI and transport params.

    We want the main code to simply call ropen(remote_file), from the class definition below, and get the remote
    file or object from whichever object or file source it came from.  We do this by prefixing the remote file
    with an identifier representing the configuration parameters for the object source.  The configuration
    parameters for each source are kept in CFG.D['Object_Sources'], or CFG.D["Dependencies"]["shared-storage"]["targets"] for
    the object store source that Caber services use to pass objects among themselves.

    When ENV_Manager processes the configuration for each source, it creates an identifier prefix (id_prefix)
    for the source and then creates a reverse mapping of prefix to source in CFG.D['Object_Sources']['revMap']

    Each source's configuration looks like:

    source_config = {"type": "s3",
                      "url": "https://s3.amazonaws.com",
                      "smartOpenPfx": "s3://",
                      "id_prefix": "s3aws",
                      "client": <object reference of client used to access the source>,
                      "transport_params": None}

    So here we strip the id_prefix from the incoming uri_as_string, lookup what source_config it belongs to,
    then, using the parameters in the source_config, call smart_open to read or write the object.  In the case
    of the default configuration, for example, the Caber shared store, the incoming uri_as_string would
    look like:
               'gss⋊⋉csi-mvp-master/x/y/z

    :param uri_as_string: Identifier for the object to access in the form <id_prefix><CFG.token><bucket>/<object_key>.
    :return: Tuple of URI with smart_open appropriate prefix, and transport parameters if needed.
    """

    # If given a uri_as_string without a Caber id_prefix, then just return it.
    if isinstance(uri_as_string, str) and not uri_as_string.count(CFG.token):
        return uri_as_string, None
    elif isinstance(uri_as_string, PosixPath):
        if not uri_as_string.parts[0].count(CFG.token):
            if uri_as_string.parts[0].count(':'):
                if uri_as_string.parts[0].startswith('file'):
                    uri_as_string = str(uri_as_string).replace(':/', ':///')
                else:
                    uri_as_string = str(uri_as_string).replace(':/', '://')
            else:
                uri_as_string = str(uri_as_string)
            return uri_as_string, None
        else:
            uri_as_string = str(uri_as_string)

    scfg = None
    if not CFG.D['Object_Sources'].get('revMap', {}):
        validated = validate_all_store_configs()
    rev_map = CFG.D['Object_Sources'].get('revMap', {})

    uri_split = uri_as_string.split(CFG.token, 1)

    target = rev_map.get(uri_split[0], None)
    if target is None:
        raise ValueError(f"[caber_parse_uri] Processsing URI input '{uri_as_string}':\n"
                         f"\tNo object-store source configuration exists for target {target}")

    # If the first letter of the identifier is 'g' then this is the global-shared source
    if uri_split[0][0].lower() == 'g':
        scfg = CFG.D["Dependencies"]["shared-storage"]["targets"].get(target) or {}
    elif uri_split[0][0].lower() == 'c':
        scfg = CFG.D['Object_Sources'].get(target) or {}
    else:
        raise ValueError(f"Malformed Caber Object_Source identifier prefix '{uri_split[0]}'.  "
                         f"The first letter of the identifier prefix must be either 'g' "
                         f"for the global shared object source, or 'c' for customer object sources")

    # Here get the relevant parameters from scfg to modify open
    scheme = scfg.get("smartOpenPfx", "file://")

    uri_as_string = uri_split[1]
    object_version = ""
    if uri_as_string.count(CFG.G["objVersSeparator"]):
        uri_as_string, object_version = uri_as_string.rsplit(CFG.G["objVersSeparator"], 1)

    uri_as_string = f"{scheme}{uri_as_string}"
    transport_params = scfg.get("transport_params", None)

    if object_version:
        if not transport_params:
            transport_params = {'version_id': object_version}
        else:
            transport_params.update({'version_id': object_version})

    return uri_as_string, transport_params


def pull_store_config(src_config='', bucket=''):
    """
    Returns the storage configuration for Caber Global Storage Source or one of the customer's
    configured object sources.

    If called with no arguments returns the global storage source configuration.
    Else checks to see if the host name is a CORI id_prefix
        if not, it looks up the CORI id_prefix using CFG.D["Object_Sources"]["hostMap"][(host + bucket)]
    to get the storage source id_prefix.
    Lastly, if the hostname is our own api-tap, then look up in the ApiMap what the upstream host is.

    Gets the store config based on 'global' for internal use
    or 'object_sources' for customer defined object sources

    If no source config given, the Global source is implied.

    Easy to get confused with teh CORI related translators in filenames.py

    """

    scfg = {}
    where = ''

    if not src_config or src_config.startswith(CFG.sspfx) or bucket == CFG.bucket:
        target = CFG.D["Dependencies"]["shared-storage"]["useTarget"]
        scfg = CFG.D["Dependencies"]["shared-storage"]["targets"].get(target) or {}
        where = CFG.sspfx
        if scfg is None:
            raise ValueError(f"[pull_store_config] Processsing src_config '{src_config}':\n"
                             f"\tGlobal storage source target {target}: Configuration does not exist")
        else:
            return scfg, where

    elif CFG.token in src_config:
        src_config = src_config.split(CFG.token)[0]

    elif src_config == AWS_S3_NAME:
        src_config = cs3idp

    loc = CFG.D["Object_Sources"].get("revMap", {}).get(src_config, '')

    if not loc:
        if src_config and bucket and bucket not in src_config:
            src_config = src_config + ':' + bucket
            bucket = ''

        where_list = [v for k, v in CFG.D["Object_Sources"].get("hostMap", {}).items() if fnmatch(src_config, k)]
        for where in where_list:
            loc = CFG.D["Object_Sources"].get("revMap", {}).get(where, '')
            if loc:
                break

    else:
        where = src_config

    if loc in CFG.D["Object_Sources"].keys():
        scfg = CFG.D["Object_Sources"][loc]
        return scfg, where

    else:
        if bucket:
            print(f'[WARNING] Configuration does not exist for object storage source {src_config}:{bucket} -> {where} ')
        return {}, None


def update_store_config(src_config=None, new_config=None):
    """
    Gets the store config based on 'global' for internal use
    or 'object_sources' for customer defined object sources

    If no source config given, the Global source is implied.
    """

    if new_config is None:
        return

    if not src_config or src_config.startswith('g'):
        target = CFG.D["Dependencies"]["shared-storage"]["useTarget"]
        CFG.D["Dependencies"]["shared-storage"]["targets"].update({target: new_config})

    else:
        src_config = src_config.split(CFG.token)[0]
        if src_config in CFG.D["Object_Sources"].get("revMap", {}).keys():
            src_config = CFG.D["Object_Sources"]["revMap"][src_config]

        if src_config in CFG.D["Object_Sources"].keys():
            CFG.D["Object_Sources"].update({src_config: new_config})

    return


# @retry(RuntimeError, total_tries=15, initial_wait=10, backoff_factor=1)
def _test_store_client(scfg, url, obst_client):
    errs = []
    stype = scfg.get('type', '')

    try:
        ctest = obst_client.list_buckets()

    except (NewConnectionError, EndpointConnectionError, ResponseParserError, ClientError) as err:
        print(f"[WARNING] {stype.capitalize()} host '{url}': {err}")
        return False
        # errs.append(f"[WARNING] {stype.capitalize()} host '{url}': {err}")
        # raise RuntimeError(f"[WARNING] {stype.capitalize()} host '{url}': retrying connection")
    else:
        return True


def match_bucket_in_source(host='', bucket='', mod='', objuct='', version=''):
    """
    Each Customer source in Object_sources

    :param host: The name of the customer source in the config
    :param bucket: The name of the bucket to test against
    :param mod: Module name of CORI - not used
    :param objuct: The object to check
    :return: True if the object in the bucket matches the config
    """

    # ENV_Manager creates a list of all buckets in the customer's configuration as a first-pass filter.
    # If there are buckets configured then they can be fnmatch patterns, so we need to check for
    # bucket matching any of the Bucket patterns.
    if not bucket:
        return False

    if not host and CFG.token in bucket:
        bs = bucket.split(CFG.token)
        host = bs[0]
        bucket = bs[1]

    first_pass_filter = any([pat for pat in CFG.D["Object_Sources"].get("Buckets", []) if fnmatch(bucket, pat)])
    if not first_pass_filter:
        return False
    elif not host and first_pass_filter:
        return True

    scfg, _ = pull_store_config(host, bucket)
    configured_buckets = scfg.get("buckets", {})
    if not configured_buckets:
        return False

    matched_bucket_prefixes = []
    # The first matching pattern wins.
    for cbkt in configured_buckets:
        bkt = list(cbkt.keys())[0]
        if fnmatch(bucket, bkt):
            matched_bucket_prefixes = cbkt[bkt].get("prefixes", [])
            break

    # Each configured bucket for a source can also have prefix patterns to match.  Even if no prefix is
    # supplied in the argument, we can still check to see if there is a wildcard match '*'.  If so, then
    # even a null prefix will match.
    #
    #        "buckets": {
    #          "newco-docs": {"prefixes": ["urn:oid:*"]} },
    #          "newco-shared": {"prefixes": ["*"]} }
    #        }
    #
    if matched_bucket_prefixes:
        if not objuct:
            # We found the bucket specified in the source but a prefix was not specified so return true
            return True
        if any([pat for pat in matched_bucket_prefixes if fnmatch(objuct, pat)]):
            return True

    return False


def create_new_bucket(new_bkt, src_config=None):
    """
    Create a new bucket in the specified object store source.  If the source is a volume then a
    directory will be created instead.

    :param new_bkt: Name of the bucket to create
    :param src_config: Creating buckets in any source other than the global storage source is for test and debugging only.
    We should NEVER create buckets in sources owned by the customer.
    :return: bucket name 'new_bkt' if sucessful, None otherwise
    """
    obj_source, where = pull_store_config()

    if obj_source.get('type', None) in ['file', 'local', 'volume']:
        bkt_path = PurePath(obj_source.get('directory', Path.home()), new_bkt)
        pbp = Path(bkt_path)
        if not pbp.exists():
            pbp.mkdir(mode=obj_source.get("mode", 754), parents=True)
            print(f"Created local directory: {bkt_path.as_uri()}")
        return bkt_path.as_uri()

    obst_client = obj_source.get('client', None)
    if not obst_client:
        return

    try:
        evb = obst_client.head_bucket(Bucket=new_bkt)
    except ClientError as err:
        if src_config == AWS_S3_NAME:
            response = obst_client.create_bucket(Bucket=new_bkt, ACL='private',
                                                 CreateBucketConfiguration={'LocationConstraint': AWS.region_name},
                                                 ObjectLockEnabledForBucket=False,
                                                 ObjectOwnership='BucketOwnerEnforced')
        else:
            response = obst_client.create_bucket(Bucket=new_bkt, ACL='private',
                                                 CreateBucketConfiguration={'LocationConstraint': AWS.region_name},
                                                 ObjectLockEnabledForBucket=False,
                                                 ObjectOwnership='BucketOwnerEnforced')

        waiter = obst_client.get_waiter('bucket_exists')
        try:
            waiter.wait(Bucket=new_bkt)
        except Exception as err:
            print(f"[CONFIG ERROR] In config: {err}")
            return
        print(f"Created new bucket '{new_bkt}' in {where} ({obj_source.get('url', '?')}) ")
    return new_bkt


@retry(ConnectionError, total_tries=15, initial_wait=10, backoff_factor=1)
def get_object_store_client(src_config=None):
    """
    Returns client only if we can connect to the source_name and the bucket is in the source_name and configured
    The configurations for the storage source targets are keyed by the hostname of the target. The hostname for
    AWS S3 is always 's3.amazonaws.com'.

    :param src_config: If no source config given, the Global source is implied.
            Else name of config in 'object_sources' for customer defined object sources
    :return: client object for given source config
    """

    scfg, where = pull_store_config(src_config)
    if type(scfg.get("client", None)).__qualname__ == 'S3':
        return scfg['client']

    adv_config = boto3config(s3={'addressing_style': 'path'},
                             user_agent=str(CFG.user_agent),
                             connect_timeout=7,
                             read_timeout=15)

    host_type = scfg.get('type', '')

    if check_s3_name(host_type) == AWS_S3_NAME:

        # Creating the client always succeeds so long as the boto3config parameters are of the correct type
        obst_client = boto3.client('s3', config=adv_config)
        scfg.update({"client": obst_client})

        for k in set(default_configs[AWS_S3_NAME].keys()).difference(scfg.keys()):
            scfg.update({k: default_configs[AWS_S3_NAME][k]})

    elif host_type.lower() in SupportedS3likeStoreTypes:
        # A non-S3 host without a url is an error since the url is needed to set up the client
        urls = first_valid(scfg.get('url'), scfg.get('urls'), oftype='any')

        if urls is None:
            scfg.update({'client': False})
            print(f"[ERROR] In config '{where}': No URL(s) configured for {host_type} Host")
            return scfg

        if isinstance(urls, str):
            urls = [urls]

        # If no client, then set one up for each URL in the URL list and quit when we find one that works
        if scfg.get('client', None) is None or scfg.get('client', '') == 'retry':
            for url in urls:
                # Creating the client always succeeds so long as the boto3config parameters are of the correct type
                obst_client = boto3.client('s3', endpoint_url=url,
                                           use_ssl=scfg.get('use_ssl', False),
                                           aws_access_key_id=scfg.get('accessKey', ''),
                                           aws_secret_access_key=scfg.get('secretKey', ''),
                                           region_name=scfg.get('region', ''),
                                           config=adv_config)

                if _test_store_client(scfg, url, obst_client):
                    scfg.update({'client': obst_client})
                    scfg.update({'url': url})
                    print(f"Instantiated {scfg.get('client_type', 'default')} client for {host_type} host")

                    if host_type.lower() == 'minio':
                        # Need to get the minio HostId to match in the logs minio puts out
                        # Boto3 has a dependency on the python requests module
                        import requests
                        from requests_aws4auth import AWS4Auth
                        auth = AWS4Auth(scfg.get('accessKey', ''), scfg.get('secretKey', ''), scfg.get('region', ''), 's3')
                        rsp = requests.get(f"{scfg['url']}/minio/admin/v3/info", auth=auth)
                        minio_id = rsp.json().get('HostId', '')
                        _build_map(minio_id, '*', _extract_bucket_list(scfg.get("buckets")),
                                   scfg.get("id_prefix").split(CFG.token)[0])
                    break

            if not scfg.get('client', None):
                raise ConnectionError(f"Failed to connect to {host_type} host")

        # For Smart-Open, all S3-like object stores get the s3:// prefix and client in the transport params
        scfg.update({"smartOpenPfx": "s3://"})
        scfg.update({"transport_params": {"client": scfg['client']}})

    elif host_type in ['local', 'file', 'volume']:
        scfg.update({"smartOpenPfx": "file://"})
        scfg.update({"client": "file"})

    update_store_config(src_config=src_config, new_config=scfg)
    return scfg.get('client', None)


# Keep at the end so all other iunctions are read in
# validate_all_store_configs()
_ = caber_patch_pathlib()




# ##################################################################################################
#  Copyright (c) 2024.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  main.py                                                                              #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

import base64
import json

from datetime import timezone, datetime
import urllib3

from .streamtap import StreamTap
from .init import CFG, post_status_message, get_primary_ip, gethostname
from .sqs_init import post2queue, get_queue
from .json_encoder import extEncoder


urllib3.disable_warnings()   # Gets rid of annoying self-signed cert errors

VERSION = "csi-genai-lib-0.1.0"
noB64encode = ['text/html', 'text/css', 'text/plain', 'text/xml', 'application/x-www-form-urlencoded',
               'application/javascript', 'application/json', 'application/xhtml+xml']

CRLF = '\n'
request_session_hold = None
queue = None
normal_ops = 0


class MainHandler:
    """
    The MainHandler class is responsible for handling API requests and responses for the GenAI Chat Library.

    This class provides methods to create a new request, update the response, and post the event to an SQS queue
    for further processing. It captures request and response data, including headers, body, and metadata, and
    prepares an SQS message with the captured information.

    Attributes:
        SUPPORTED_METHODS (tuple): A tuple of supported HTTP methods.
        reqB64encode (bool): Flag indicating whether to base64 encode the request body.
        request (dict): Dictionary containing the request information.
        isStream (bool): Flag indicating if the request is a stream.
        dest_host (str): Destination host for the request.
        path (str): API path for the request.
        req_timeout (int): Request timeout in seconds.
        default_method (str): Default HTTP method for the request.
        allow_redirects (bool): Flag indicating whether to allow redirects.
        auth (dict): Dictionary containing user authentication information.
        sqs_message (dict): Dictionary representing the SQS message to be sent.
        updata (str): Request body data.
        apirsp (object): API response object.
        apimap (dict): API mapping configuration.
        dest_url (str): Destination URL for the request.
        req_id (str): Unique request ID.
        req_tap (StreamTap): StreamTap object for capturing request data.
        rsp_tap (StreamTap): StreamTap object for capturing response data.
        session (str): User session identifier.
        req_phdrs (dict): Dictionary containing request headers.
        rsp_phdrs (dict): Dictionary containing response headers.
        query (str): Query string for the request.
        user_session (str): User session identifier.
        user_host (str): User hostname or IP address.
        port (int): Port number for the request.
        mime (str): MIME type for the request and response.
        ready_to_post (int): Flag indicating the readiness to post the event.

    Methods:
        __init__(**kwargs): Initializes a new instance of the MainHandler class.
        new_request(**kwargs): Creates a new request with the provided parameters.
        post_event(): Posts the captured event to an SQS queue.
        update_response(**kwargs): Updates the response with the provided parameters.
    """

    SUPPORTED_METHODS = ('GET', 'HEAD', 'POST', 'DELETE', 'PUT', 'PATCH', 'OPTIONS',
                         'PROPFIND', 'MKCOL', 'MOVE', 'LOCK', 'UNLOCK', 'PROPPATCH')
    reqB64encode = False
    request = {}
    isStream = False
    dest_host = gethostname()
    path = 'chat-post'
    req_timeout = 10
    default_method = "POST"
    allow_redirects = True
    auth = {"user.id": "noAuth", "user.type": "basic"}
    _sqs_message = {}
    updata = ''
    apirsp = None
    apimap = None
    dest_url = ''
    req_id = 'should be a uuid here'
    req_tap = None
    rsp_tap = None
    session = ''
    req_phdrs = {}
    rsp_phdrs = {}
    query = None
    user_session = None
    user_host = "unknown.nosession"
    port = 80
    mime = 'application/json; charset=utf-8'
    ready_to_post = 0

    def __init__(self, **kwargs) -> None:
        self._sqs_message = {"R.ID": "", "event.time": datetime.now(timezone.utc).isoformat(), "R.VERSION": VERSION,
                             'R.API': '', 'C.REQ': {}, 'C.RSP': {}}
        self.ready_to_post = 0
        try:
            post_status_message(stage="3.0", message="Entering normal operation", status="green")
        except Exception as e:
            pass

    @property
    def sqs_message(self) -> dict:
        if self.ready_to_post == 0:
            return {}
        return self._sqs_message    

    def convert_updata(self, charset='utf-8') -> None:
        if isinstance(self.updata, (dict, list)):
            self.updata = json.dumps(self.updata)
            self.updata = self.updata.encode(charset)
        elif isinstance(self.updata, str):
            self.updata = base64.b64decode(self.updata) if self.reqB64encode else self.updata.encode(charset)
        elif not isinstance(self.updata, bytes):
            self.updata = str(self.updata).encode(charset)

    def new_request(self, **kwargs) -> None:
        global normal_ops
        self.updata = kwargs.get('context') or ''
        self.reqB64encode = kwargs.get('base64') or False
        self.dest_host = kwargs.get('hostname') or gethostname()
        self.req_timeout = kwargs.get("req_timeout") or 10
        self.default_method = kwargs.get("default_method") or "POST"
        self.auth = {"user.id": kwargs.get("user_id", "noAuth"), "user.type": kwargs.get("user_type", "basic")}
        self.user_host = kwargs.get('user_hostname_or_ip') or "unknown." + str(kwargs.get('user_session', 'nosession'))
        self.user_session = kwargs.get('user_session') or None
        self.path = kwargs.get('api_name') or 'chat-post'
        self.mime = kwargs.get('mime_type') or 'application/json; charset=utf-8'
        self.query = kwargs.get('query') or None
        self.req_id = CFG.new_id()
        self.req_tap = None
        self.rsp_tap = None
        self.req_phdrs = {}
        self.rsp_phdrs = {}
        self.port = 80

        self.path = self.path.strip('/')
        if isinstance(self.dest_host, str) and self.dest_host.endswith('/'):  # Drop trailing '/'
            self.dest_host = self.dest_host[:-1]

        psswd = base64.b64encode(f'{self.auth.get("user.id", "noAuth")}:abc123'.encode('utf-8')).decode('utf-8')

        self._sqs_message = {
            "R.ID": self.req_id,
            "event.id": self.req_id,
            "event.time": datetime.now(timezone.utc).isoformat(),
            "R.VERSION": VERSION,
            'R.API': f'{self.dest_host}/{self.path}',
            'api.host.name': self.dest_host,
            'api.path': self.path,
            "user.id": kwargs.get("user_id", "noAuth"),
            "user.type": kwargs.get("user_type", "basic"),
            'user.from.host.ip': self.user_host,
            'user.from.host.name': self.user_host,
            'C.REQ': {
                "method": self.default_method,
                "host": self.dest_host,
                "path": self.path,
                "base64": False,
                "sha256": "",
                "body": "",
                "headers": {
                    "Host": self.dest_host,
                    "User-Agent": VERSION,
                    'Content-Type': self.mime,
                    "Authorization": f"{self.auth.get('user.type', 'basic').capitalize()} {psswd}",
                },
            },
            'C.RSP': {
                "code": 200,
                "base64": False,
                "sha256": "",
                "body": "",
                "headers": {},
            }
        }

        normal_ops += 1
        if CFG.post_log_now:
            post_status_message(stage="3.1", message=f"Processing API request {normal_ops}", status="green")

        if self.query is not None:
            self._sqs_message.update({'R.ARGS': f'{self.query}'})
            self._sqs_message['C.REQ'].update({'query': f'{self.query}'})

        if self.user_session is not None:
            self._sqs_message.update({'user.session': self.user_session})

        req_blen = self.request.get('headers', {}).get('Content-Length') or len(self.request.get('body', ''))
        req_type = self.mime
        if isinstance(req_type, str):
            req_type = req_type.split('; ')
        if len(req_type) == 1:
            req_type.append('charset=utf-8')
        charset = req_type[1].split('=')[-1] or 'utf-8'

        self.reqB64encode = False if req_type[0] in noB64encode else True
        self.convert_updata(charset)

        req_blen = len(self.updata)
        xuid = f'REQ-{self.req_id[-16:]}'
        print(f"[{self.req_id[-12:]}] RQST Buffered data: Type={req_type}  Length={req_blen}")
        self.req_tap = StreamTap(self.updata, chunk_size=req_blen, uid=xuid)
        self.req_tap.read()

        # Check if configured to post to SQS for the current status code
        # codeRe is regex pattern for matching status codes
        if self.req_tap is not None:
            if not self.req_tap.closed:
                self.req_tap.close()
            if self.req_tap.result is not None:
                self._sqs_message['C.REQ'].update({"sha256": self.req_tap.sha256})
                if isinstance(self.req_tap.result, str):
                    self._sqs_message['C.REQ'].update({"body": self.req_tap.result})
                elif self.reqB64encode:
                    self._sqs_message['C.REQ'].update({"body": base64.b64encode(self.req_tap.result).decode('utf-8')})
                    self._sqs_message['C.REQ'].update({"base64": True})
                else:
                    self._sqs_message['C.REQ'].update({"body": self.req_tap.result.decode('utf-8')})
                pmsg = self._sqs_message["C.REQ"]["body"][0:80].replace(CRLF, '')
                print(f'[{self.req_id[-12:]}] RQST Data tap result => {pmsg} self.req_tap.result={self.req_tap.result}')

        self.ready_to_post = 1

    def post_event(self) -> None:
        if self.ready_to_post != 2:
            raise RuntimeError("Not ready to post:  You must call new_request(), update_response() first")

        self.ready_to_post = 0

        message = json.dumps(self._sqs_message, cls=extEncoder)
        if len(message) < 262143:
            sqs_queue = get_queue("APIshdwQueue")
            # Send message to SQS queue for consumption by Process_API
            sqs_resp = post2queue(sqs_queue, message)

            print(f"[{self.req_id[-12:]}] DONE: Tap posted to "
                  f"{sqs_queue.split('/')[-1]} "
                  f"({sqs_resp['MessageId']})")
        else:
            print(
                f"[{self.req_id[-12:]}] DONE: No post to SQS: Message too big - {len(message)} > {CFG.me('maxSqsLen')}")

    def update_response(self, **kwargs) -> None:
        if self.ready_to_post != 1:
            raise RuntimeError("Not ready for response:  You must call new_request() first")

        self.apirsp = kwargs.get('response') or None
        rsp_type = kwargs.get('mime_type') or 'application/json; charset=utf-8'

        data = True if self.apirsp is not None else False
        alen = len(self.apirsp) if data else 0
        rsp_params = {
            "code": 200,
            "base64": False,
            "sha256": "",
            "body": "",
            "headers": {
                'Content-Type': self.mime,
            },
        }

        if isinstance(rsp_type, str):
            rsp_type = rsp_type.split('; ')
        if len(rsp_type) == 1:
            rsp_type.append('charset=utf-8')
        charset = rsp_type[1].split('=')[-1] or 'utf-8'
        rspB64encode = False if rsp_type[0] in noB64encode else True

        xuid = f'RSP-{self.req_id[-16:]}'
        if int(alen) <= 0:
            print(f"[{self.req_id[-12:]}] RESP NO body")
        else:
            print(f"[{self.req_id[-12:]}] RESP Body: Type={rsp_type}  Length={alen}")
            self.rsp_tap = StreamTap(self.apirsp, chunk_size=alen, uid=xuid)
            # self.req_tap.read()

            if not self.rsp_tap.closed:
                self.rsp_tap.close()
            if self.rsp_tap.result is not None:
                rsp_params.update({"sha256": self.rsp_tap.sha256})
                if isinstance(self.rsp_tap.result, str):
                    rsp_params.update({"body": self.rsp_tap.result})
                elif rspB64encode:
                    rsp_params.update({"body": base64.b64encode(self.rsp_tap.result).decode(charset)})
                    rsp_params.update({"base64": True})
                else:
                    rsp_params.update({"body": self.rsp_tap.result.decode(charset)})
                pmsg = self._sqs_message["C.RSP"]["body"][0:80].replace(CRLF, '')
                print(f'[{self.req_id[-12:]}] RESP Data tap result => {pmsg}')

        self._sqs_message.update({'C.RSP': rsp_params})
        self.ready_to_post = 2



# ##################################################################################################
#  Copyright (c) 2023.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  If used, attribution of open-source code is included where required by original author          #
#                                                                                                  #
#  Filename:  supabase.py                                                                          #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

import requests
import os
from datetime import datetime


SUPABASE_URL = os.getenv("CSI_SUPABASE_URL")
SUPABASE_KEY = os.getenv("CSI_SUPABASE_API_KEY")
SUPABASE_LOGS_TABLE = os.getenv("CSI_DEPLOYMENT_LOGS_TABLE")

HEADERS = {
    "apikey": SUPABASE_KEY,
    "Authorization": f"Bearer {SUPABASE_KEY}",
    "Content-Type": "application/json",
}


def post_status_to_supabase(log_data):
    if not SUPABASE_URL or not SUPABASE_KEY or not SUPABASE_LOGS_TABLE:
        print("Supabase environment variables not set. Skipping log post.")
        return False
    if not SUPABASE_URL.startswith("https://"):
        print("Supabase URL must start with https://. Skipping log post.")
        return False
    if not isinstance(log_data, dict):
        print("Log data must be a dictionary. Skipping log post.")
        return False

    # Force the timestamp to be UTC
    log_data["timestamp"] = datetime.utcnow().isoformat()

    must_have_keys = ["deployment_id", "stage", "category", "message", "group", "timestamp", "status"]
    good_keys = ["deployment_id", "stage", "category", "message", "group", "timestamp", "status", "progress_bar", "extra"]
    filtered_log_data = {k: v for k, v in log_data.items() if v and k in good_keys}

    check = set(must_have_keys).difference(set(filtered_log_data.keys()))

    if len(check):
        print(f"Log data missing the following keys: {check}.")

    try:
        # print(f"Sending the following to Supabase: {filtered_log_data}")
        response = requests.post(
            f"{SUPABASE_URL}/rest/v1/{SUPABASE_LOGS_TABLE}",
            headers=HEADERS,
            json=filtered_log_data
        )
    except Exception as e:
        print(f"Failed to post log. Exception: {e}")
        return False

    if response.status_code != 201:
        print(f"Failed to post log. Status Code: {response.status_code}. Response: {response.text}")
        return False

    return True



# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  aws_init.py                                                                          #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

"""
This module attempts to abstract away the complexities of dealing with multiple types of AWS_S3-like object
store services.  Even though https://pypi.org/project/smart-open/ gives us the ability to read and write objects
and files from many types of sources, there are source-dependent quirks that have to be dealt with.

For example, although Min.io is supposed to be completely S3 compatible, it requires a username:password
combo to be passed and the S3 client must have the URL of the Minio host defined.

The other quirk is that for use internal to Caber services we can use AWS_S3 or another source, and customers
can define multiple types of hosts simultaneously.  All of these we keep track of here.

"""

import os
import json
import pytz
import boto3
from uuid import uuid4
from datetime import datetime
from time import time

from boto3 import Session
from .json_encoder import extEncoder
from urllib.parse import urlparse
from smart_open.smart_open_lib import patch_pathlib
from botocore.config import Config as boto3config
from botocore.exceptions import ClientError
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session

# THIS MODULE IS IMPORTED INTO Common.init so we cannot import CFG here!
#   NO -->  from ..Common.init import CFG

# ############# INCLUDE HERE THE NAMES OF TESTED OBJECT STORE TYPES ################ #
#   These form the possible choices for:
#          CFG.D['Object_Sources'][<host-name>]['type']
#          CFG.D["Dependencies"]["shared-storage"]["targets"][<host-name>]['type']
#
AWS_S3_SHORT_NAME = 's3aws'
AWS_S3_NAME = 's3.amazonaws.com'
SupportedS3likeStoreTypes = [AWS_S3_NAME, 'minio']
# ################################################################################## #

_ = patch_pathlib()  # replace `Path.open` with `smart_open.open`



class RefreshableBotoSession:
    """
    From https://stackoverflow.com/questions/63724485/how-to-refresh-the-boto3-credentials-when-python-script-is-running-indefinitely
    Boto Helper class which lets us create a refreshable session so that we can cache the client or resource.

    Usage
    -----
    session = RefreshableBotoSession().refreshable_session()

    client = session.client("s3") # we now can cache this client object without worrying about expiring credentials
    """

    def __init__(
        self,
        region_name: str = None,
        profile_name: str = None,
        access_key: str = None,
        secret_key: str = None,
        sts_arn: str = None,
        session_name: str = None,
        session_ttl: int = 1500
    ):
        """
        Initialize `RefreshableBotoSession`

        Parameters
        ----------
        region_name : str (optional)
            Default region when creating a new connection.

        profile_name : str (optional)
            The name of a profile to use.

        sts_arn : str (optional)
            The role arn to sts before creating a session.

        session_name : str (optional)
            An identifier for the assumed role session. (required when `sts_arn` is given)

        session_ttl : int (optional)
            An integer number to set the TTL for each session. Beyond this session, it will renew the token.
            25 minutes by default which is before the default role expiration of 1 hour
        """

        self.region_name = region_name
        self.profile_name = profile_name
        self.sts_arn = os.getenv('CSI_TENANT_ROLE_ARN')
        self.ext_id = os.getenv('CSI_ROLE_EXTERNAL_ID')
        self.session_name = session_name or uuid4().hex
        self.session_ttl = session_ttl
        self.access_key = os.environ.get('CSI_AWS_ACCESS_KEY')
        self.secret_key = os.environ.get('CSI_AWS_SECRET_KEY')
        self.session = None

    def __get_new_session(self):
        """
        Get session credentials
        """
        if self.profile_name:
            print(f"RefreshableBotoSession: Using session credentials from profile {self.profile_name}")
            self.session = Session(region_name=self.region_name, profile_name=self.profile_name)
        elif self.access_key and self.secret_key:
            print(f"RefreshableBotoSession: Using session credentials from environment variables")
            self.session = Session(region_name=self.region_name, aws_access_key_id=self.access_key, aws_secret_access_key=self.secret_key)
        else:
            print(f"RefreshableBotoSession: Using default session credentials")
            self.session = Session(region_name=self.region_name)

    def __get_session_credentials(self):
        if not self.session:
            self.__get_new_session()
        credentials = {}
        response = {}
        # if sts_arn is given, get credential by assuming the given role
        if self.sts_arn:
            print(f"RefreshableBotoSession: Assuming role from environment variables")
            try:
                sts_client = self.session.client(service_name="sts", config=aws_session_config)
                assumed_role_object = sts_client.assume_role(
                    RoleArn=self.sts_arn,
                    ExternalId=self.ext_id,
                    RoleSessionName=self.session_name,
                    DurationSeconds=self.session_ttl,
                )
                response = assumed_role_object.get("Credentials")
            except ClientError as err:
                print(f"[WARNING] assume_customer_role: Exception assuming Cross Account Role ${err}")
            else:
                credentials = {
                    "access_key": response.get("AccessKeyId"),
                    "secret_key": response.get("SecretAccessKey"),
                    "token": response.get("SessionToken"),
                    "expiry_time": response.get("Expiration").isoformat(),
                }
                print(f"RefreshableBotoSession: Assuming role successful. Session expiration: {credentials['expiry_time']}")

        if not credentials:
            session_credentials = self.session.get_credentials().get_frozen_credentials()
            credentials = {
                "access_key": session_credentials.access_key,
                "secret_key": session_credentials.secret_key,
                "token": session_credentials.token,
                "expiry_time": datetime.fromtimestamp(time() + self.session_ttl).replace(tzinfo=pytz.utc).isoformat(),
            }
            print(f"RefreshableBotoSession: Assuming role failed. Using base role for session. Session expiration: {credentials['expiry_time']}")

        os.environ['AWS_ACCESS_KEY_ID'] = credentials.get("access_key")
        os.environ['AWS_SECRET_ACCESS_KEY'] = credentials.get("secret_key")
        os.environ['AWS_SESSION_TOKEN'] = credentials.get("token")
        return credentials

    def refreshable_session(self) -> Session:
        """
        Get refreshable boto3 session.
        """
        # Get refreshable credentials
        initial_credentials = self.__get_session_credentials()
        refreshable_credentials = RefreshableCredentials.create_from_metadata(
            metadata=initial_credentials,
            refresh_using=self.__get_session_credentials,
            method="auto",
        )

        # attach refreshable credentials current session
        autorefresh_session = boto3.Session()
        autorefresh_session._session.set_credentials(refreshable_credentials)
        autorefresh_session._session.set_config_variable("region", self.region_name)
        autorefresh_session._session.set_config_variable("user_agent", user_agent)

        return autorefresh_session


def get_aws_regions():
    global AWS_REGIONS

    if AWS_REGIONS is None:
        client = new_aws_client('ec2')
        try:
            AWS_REGIONS = [region['RegionName'] for region in client.describe_regions()['Regions']]
        except ClientError as err:
            print(f"Using default for AWS Regions list. Add 'ec2:DescribeRegions' permissions.")
            AWS_REGIONS = ['ap-south-1', 'eu-north-1', 'eu-west-3', 'eu-west-2', 'eu-west-1',
                           'ap-northeast-3', 'ap-northeast-2', 'ap-northeast-1', 'ca-central-1',
                           'sa-east-1', 'ap-southeast-1', 'ap-southeast-2', 'eu-central-1',
                           'us-east-1', 'us-east-2', 'us-west-1', 'us-west-2']
    return AWS_REGIONS


def check_s3_name(name):
    if 's3-scanner' in name or 's3_scanner' in name:
        return name
    s3list = ['s3', AWS_S3_SHORT_NAME, 'awss3', 'amazonawss3', 'amazons3', AWS_S3_NAME]
    if isinstance(name, str) and name.lower() in s3list or any([1 for x in s3list if x in name.lower()]):
        return AWS_S3_NAME
    else:
        return name


def url_to_from_arn(cfg, name, acct=None, region=None):
    fs = name.find('/')
    nc = name[:fs].count(':')
    if fs > 0 and nc == 1:   # One colon before first slash means it's a URL
        if acct is None:
            acct = aws_account_id
        if region is None:
            region = AWS.region_name
        url = urlparse(name)
        obj_type = cfg.D["Object_Sources"].get(check_s3_name(url.hostname), {}).get('type', '')
        arn = f"arn:{obj_type}:{url.scheme}:{region}:{acct}:{url.path}"
        return arn
    elif nc == 5:
        arn = name.split(':', 5)
        url = arn[2] + '://' + arn[-1]
        return url
    else:
        return name


notification_bucket_policy = '{"Version": "2012-10-17", "Statement": [{"Sid": "AWSCloudTrailAclCheck20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "s3:GetBucketAcl", "Resource": "%bk_arn%"}, {"Sid": "AWSCloudTrailWrite20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "s3:PutObject", "Resource": "%ct_arn%", "Condition": {"StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}}}, {"Sid": "AWSLogDeliveryWrite", "Effect": "Allow", "Principal": {"Service": "delivery.logs.amazonaws.com"}, "Action": "s3:PutObject", "Resource": "%cw_arn%", "Condition": {"StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}}}, {"Sid": "AWSLogDeliveryAclCheck", "Effect": "Allow", "Principal": {"Service": "delivery.logs.amazonaws.com"}, "Action": "s3:GetBucketAcl", "Resource": "%bk_arn%"}]}'
notification_topic_policy = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__default_statement_ID", "Effect": "Allow", "Principal": {"AWS": "*"}, "Action": ["SNS:GetTopicAttributes", "SNS:SetTopicAttributes", "SNS:AddPermission", "SNS:RemovePermission", "SNS:DeleteTopic", "SNS:Subscribe", "SNS:ListSubscriptionsByTopic", "SNS:Publish", "SNS:Receive"], "Resource": "%topic_arn%", "Condition": {"StringEquals": {"AWS:SourceOwner": "%aws_id%"}}}, {"Sid": "AWSCloudTrailSNSPolicy20150319", "Effect": "Allow", "Principal": {"Service": "cloudtrail.amazonaws.com"}, "Action": "SNS:Publish", "Resource": "%topic_arn%"}]}'
event_queue_policy_sns = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__owner_statement", "Effect": "Allow"," Principal": {"AWS":"*"},"Action": "SQS:*", "Resource": "%queue_arn%"}, {"Sid": "topic-subscription-%topic_arn%", "Effect": "Allow", "Principal": {"AWS":"*"}, "Action": "SQS:*", "Resource": "%queue_arn%", "Condition": {"ArnLike": {"aws:SourceArn": "%topic_arn%"}}}]}'
event_queue_policy = '{"Version": "2008-10-17", "Id": "__default_policy_ID", "Statement": [{"Sid": "__owner_statement", "Effect": "Allow"," Principal": {"AWS":"*"},"Action": "SQS:*", "Resource": "%queue_arn%"}]}'


def create_event_notification_bucket(cfg, event_bkt="csi-cloudtrail-events-"):
    if event_bkt.endswith('-'):
        event_bkt += aws_credentials.access_key

    evb = "create_new_bucket(cfg, event_bkt)"

    bk_arn = f"arn:aws:s3:::{event_bkt}"
    ct_arn = f"arn:aws:s3:::{event_bkt}/AWSLogs/{aws_account_id}/*"
    cw_arn = f"arn:aws:s3:::{event_bkt}/flow-logs/AWSLogs/{aws_account_id}/*"

    nb_policy = notification_bucket_policy.replace('%bk_arn%', bk_arn)\
                                          .replace('%ct_arn%', ct_arn)\
                                          .replace('%cw_arn%', cw_arn)

    resp = S3C.put_bucket_policy(Bucket=event_bkt, Policy=nb_policy)
    print(f"Added bucket policy to: {event_bkt}")
    return event_bkt


def create_sns_notification_topic(event_sns="csi-new-logfile-event-"):
    if event_sns.endswith('-'):
        event_sns += aws_credentials.access_key

    topic_arn = f"arn:aws:sns:{AWS.region_name}:{aws_account_id}:{event_sns}"
    lg_policy = notification_topic_policy.replace('%topic_arn%', topic_arn)\
                                         .replace('%aws_id%', aws_account_id)

    sns = new_aws_client('sns')

    try:
        sns_attr = sns.get_topic_attributes(TopicArn=topic_arn)
        sqs_q_name = create_sqs_q(sns_topic_arn=topic_arn)
    except Exception as err:
        sns_resp = sns.create_topic(Name=event_sns,
                                    Attributes={'Policy': lg_policy})
        sns_attr = sns.get_topic_attributes(TopicArn=topic_arn)
        sqs_q_name = create_sqs_q(sns_topic_arn=topic_arn)
        print(f"Created SNS topic: {event_sns}")
        return topic_arn
    else:
        return topic_arn


def create_sqs_q(sqs_queue="csi-s3-data-events-", sns_topic_arn=None):
    if sqs_queue.endswith('-'):
        sqs_queue += aws_credentials.access_key

    queue_arn = f"arn:aws:sqs:{AWS.region_name}:{aws_account_id}:{sqs_queue}"
    if sns_topic_arn is None:
        sq_policy = event_queue_policy.replace('%queue_arn%', queue_arn) \
                                      .replace('%aws_id%', aws_account_id)
    else:
        sq_policy = event_queue_policy_sns.replace('%queue_arn%', queue_arn) \
                                          .replace('%topic_arn%', sns_topic_arn)\
                                          .replace('%aws_id%', aws_account_id)

    sqs = new_aws_client('sqs')

    try:
        sqs_resp = sqs.set_queue_attributes(QueueName=sqs_queue,
                                            Attributes={'Policy': sq_policy})
    except Exception as err:
        sqs_resp = sqs.create_queue(QueueName=sqs_queue,
                                    Attributes={'Policy': sq_policy})
        sqs_attr = sqs.get_queue_attributes(QueueName=sqs_queue)
        print(f"Created SQS queue: {sqs_queue}")
        return sqs_queue
    else:
        return sqs_queue



from datetime import datetime, timezone


def post_status_to_cloudwatch(status):
    """
    Post a message to the specified AWS CloudWatch log group.

    Example Usage
    response = post_status_to_cloudwatch("This is a test message", "TestLogGroup")
    print(response)

    FORMAT_STRING: '[CSI-STATUS] {"group": "codebuild", "stage": "%s.%s", "message": "%s %s"}\n'

    :param status: The message to post
    """
    log_group = os.environ.get("CSI_STATUS_LOG_GROUP")

    if not log_group:
        # print(f"[WARNING] post_status_to_cloudwatch: Failed 'CSI_STATUS_LOG_GROUP' not set.")
        return False

    message = status.copy()

    sequence_token = None
    external_id = os.getenv('CSI_ROLE_EXTERNAL_ID')

    # timestamp the message
    timestamp = int(datetime.now(tz=timezone.utc).timestamp() * 1000)  # AWS uses milliseconds

    if isinstance(message, dict):
        message |= {'timestamp': timestamp, 'external_id': external_id}
        message = f"[CSI-STATUS] {json.dumps(message, cls=extEncoder)}"
        event = {'timestamp': timestamp, 'message': message}
    else:
        event = {'timestamp': timestamp, 'message': message}

    # Try to get the log streams
    try:
        streams = LOGS.describe_log_streams(logGroupName=log_group, limit=1)
    except LOGS.exceptions.ResourceNotFoundException:
        # If log group not found, create it and retry
        LOGS.create_log_group(logGroupName=log_group)
        streams = LOGS.describe_log_streams(logGroupName=log_group, limit=1)

    # If there are no streams, or the last stream is full, create a new stream
    if not streams['logStreams'] or 'uploadSequenceToken' not in streams['logStreams'][0]:
        stream_name = f'status-{int(datetime.now(tz=timezone.utc).timestamp())}'
        LOGS.create_log_stream(logGroupName=log_group, logStreamName=stream_name)
    else:
        # Use the existing stream
        stream = streams['logStreams'][0]
        stream_name = stream['logStreamName']
        sequence_token = stream.get('uploadSequenceToken')

    if sequence_token:
        # If sequence token exists, include it in the put request
        response = LOGS.put_log_events(
            logGroupName=log_group,
            logStreamName=stream_name,
            logEvents=[event],
            sequenceToken=sequence_token
        )
    else:
        # If not, exclude the sequence token
        response = LOGS.put_log_events(
            logGroupName=log_group,
            logStreamName=stream_name,
            logEvents=[event]
        )

    return response


'''
aws cloudtrail create-trail — name log-archive-trail — s3-bucket-name $s3_bucket_name — region $TARGET_REGION
aws cloudtrail put-event-selectors — trail-name log-archive-trail — region $TARGET_REGION — event-selectors ‘[{“ReadWriteType”: “All”,”IncludeManagementEvents”: true,”DataResources”: [{“Type”:”AWS::S3::Object”, “Values”: [“arn:aws:s3:::”]}]}]’
aws cloudtrail start-logging — name log-archive-trail
'''


def create_s3_cloudtrail(cfg, buckets, event_bkt="csi-cloudtrail-events-", event_sns="csi-new-logfile-event-"):

    ev_bkt_name = create_event_notification_bucket(cfg, event_bkt)
    sns_topic_arn = create_sns_notification_topic(event_sns)

    ct = new_aws_client('cloudtrail')
    trail_name = f'csi-s3-trail-{aws_credentials.access_key}'

    trail = ct.create_trail(Name=trail_name,
                            S3BucketName=ev_bkt_name,
                            SnsTopicName=sns_topic_arn,
                            IncludeGlobalServiceEvents=False,
                            IsMultiRegionTrail=True,
                            EnableLogFileValidation=False)

    ct_event_selectors = '[{"ReadWriteType":"All", "IncludeManagementEvents":false, "DataResources": [{"Type": "AWS::S3::Object", "Values": %bkt_arns%}], "ExcludeManagementEventSources":[]}]'

    bkt_arns = []
    for bkt in buckets:
        if bkt.count('arn'):
            if bkt.endswith('/'):
                bkt_arns.append(bkt)
            else:
                bkt_arns.append(bkt + '/')
        else:
            if bkt.endswith('/'):
                bkt_arns.append(f"arn:aws:s3:::{bkt}")
            else:
                bkt_arns.append(f"arn:aws:s3:::{bkt}/")

    if bkt_arns:
        bkts = json.dumps(bkt_arns)
        ct_event_selectors.replace('%bkt_arns%', bkts)
        ct.put_event_selectors(TrailName=trail_name, EventSelectors=json.loads(ct_event_selectors))
        ct.start_logging(Name=trail_name)

    return


def assume_customer_role():
    global AWS
    global aws_credentials
    global aws_account_id

    # Fetch the credentials from environment variables
    access_key = os.environ.get('CSI_AWS_ACCESS_KEY')
    secret_key = os.environ.get('CSI_AWS_SECRET_KEY')

    if not access_key or not secret_key:
        print(f"[WARNING] assume_customer_role: Failed 'CSI_AWS_ACCESS_KEY' and/or 'CSI_AWS_SECRET_KEY' not set.")
        return False

    try:
        # Create an STS client using the credentials from SecretsManager
        sts_client = boto3.client(
            'sts',
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key
        )
    except ClientError as err:
        print(f"[WARNING] assume_customer_role: Exception creating sts_client ${err}")
        return False

    cross_account_role_arn = os.environ.get('CSI_TENANT_ROLE_ARN')
    cross_account_role_external_id = os.environ.get('CSI_ROLE_EXTERNAL_ID')
    if not cross_account_role_arn or not cross_account_role_external_id:
        print(f"[WARNING] assume_customer_role: Failed 'CSI_TENANT_ROLE_ARN' and/or 'CSI_ROLE_EXTERNAL_ID' not set.")
        return False

    try:
        assumed_role_object = sts_client.assume_role(
            RoleArn=cross_account_role_arn,
            RoleSessionName="CrossAccountAssumptionSession",
            ExternalId=cross_account_role_external_id
        )
    except ClientError as err:
        print(f"[WARNING] assume_customer_role: Exception assuming Cross Account Role ${err}")
        return False

    # Credentials from the assumed role
    assumed_credentials = assumed_role_object.get('Credentials', {})
    if not {'AccessKeyId', 'SecretAccessKey', 'SessionToken'}.intersection(assumed_credentials.keys()):
        print(f"[WARNING] assume_customer_role: STS returned incomplete credentials when assuming the Cross Account Role")
        return False

    os.environ['AWS_ACCESS_KEY_ID'] = assumed_credentials['AccessKeyId']
    os.environ['AWS_SECRET_ACCESS_KEY'] = assumed_credentials['SecretAccessKey']
    os.environ['AWS_SESSION_TOKEN'] = assumed_credentials['SessionToken']

    try:
        AWS = boto3.session.Session()
        aws_credentials = AWS.get_credentials()
        aws_account_id = AWS.client('sts').get_caller_identity().get('Account', '')
    except Exception as err:
        print(f"[WARNING] assume_customer_role: After assuming cross account role sts.get_caller_identity() failed")
        return False

    return True


#
# TRY TO ASSUME THE CUSOMTER ROLE AN IF THAT FAILS TRY TO USE WHATEVER CREDENTIALS ARE AVAILABLE
# INHERITED FROM THE CONTAINER OR THE INSTANCE IT RUNS ON.
#
AWS = None
aws_credentials = {}
aws_account_id = ''

try:
    AWS = RefreshableBotoSession().refreshable_session()  # boto3.session.Session()
    aws_credentials = AWS.get_credentials()
    aws_account_id = AWS.client('sts').get_caller_identity().get('Account', '')
except Exception as err:
    print(f"[WARNING] aws_init: RefreshableBotoSession Failed. Falling back to assume_customer_role")
    if not assume_customer_role():
        print(f"[WARNING] aws_init: assume_customer_role Failed. Falling back to boto3.session.Session()")
        try:
            AWS = boto3.session.Session()
            aws_credentials = AWS.get_credentials()
            aws_account_id = AWS.client('sts').get_caller_identity().get('Account', '')
        except Exception as err:
            print(f"[WARNING] aws_init: Exception on initial call to sts.get_caller_identity and failed to assume customer role.  This is bad...\n{err} ")
            aws_account_id = ''

user_agent = f"csi.{__package__}.{aws_credentials.access_key}"
aws_session_config = boto3config(
    user_agent=user_agent,
    connect_timeout=7,
    read_timeout=20
)

if os.getenv("AWS_ACCESS_KEY_ID", None) is None:
    os.environ["AWS_ACCESS_KEY_ID"] = aws_credentials.access_key

if os.getenv("AWS_SECRET_ACCESS_KEY", None) is None:
    os.environ["AWS_SECRET_ACCESS_KEY"] = aws_credentials.secret_key

if os.getenv("AWS_ACCOUNT_ID", None) is None:
    os.environ["AWS_ACCOUNT_ID"] = aws_account_id

if os.getenv("AWS_DEFAULT_REGION", None) is None:
    if AWS.region_name is not None:
        os.environ["AWS_DEFAULT_REGION"] = AWS.region_name
    else:
        os.environ["AWS_DEFAULT_REGION"] = 'us-west-1'


def new_aws_client(name='s3'):
    return AWS.client(name, config=aws_session_config)


def new_aws_resource(name='s3'):
    return AWS.resource(name, config=aws_session_config)


S3C = new_aws_client()
S3R = new_aws_resource()
LOGS = new_aws_client('logs')

Default_S3 = {"type": "s3",
              "url": "https://s3.amazonaws.com",
              "smartOpenPfx": "s3://",
              "id_prefix": "s3aws",
              "client": S3C,
              "transport_params": None}

AWS_REGIONS = None


# ##################################################################################################
#  Copyright (c) 2023.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  If used, attribution of open-source code is included where required by original author          #
#                                                                                                  #
#  Filename:  retry.py                                                                             #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  retry.py                                                                             #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

#Copyright 2021 Fabian Bosler

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from functools import wraps
import time
import logging
import random
import os

logger = logging.getLogger(__name__)


def retry(exceptions, total_tries=6, initial_wait=0.1, backoff_factor=2.5, logger=None, no_print=False, fail_func=None):
    """
    calling the decorated function applying an exponential backoff.
    Args:
        exceptions: Exception(s) that trigger a retry, can be a tuple
        total_tries: Total tries
        initial_wait: Time to first retry
        backoff_factor: Backoff multiplier (e.g. value of 2 will double the delay each retry).
        logger: logger to be used, if none specified print
        no_print: Don't print retrying messages until ready to give up
    """
    def retry_decorator(f):
        @wraps(f)
        def func_with_retries(*args, **kwargs):
            _tries, _delay = total_tries + 1, initial_wait
            while _tries > 1:
                try:
                    if (total_tries + 2 - _tries) > 1 and not no_print:
                        log(f'<retry #{total_tries + 2 - _tries}>', logger)
                    return f(*args, **kwargs)
                except exceptions as e:
                    _tries -= 1
                    print_args = args if args else 'no args'
                    if _tries == 1:
                        if fail_func is not None:
                            try:
                                fail_func()
                            except Exception as e:
                                log(f'Retry: Failed to run fail_func: {e}', logger)
                            else:
                                log(f'Retry: fail_func ran successfully', logger)
                                time.sleep(10)
                                _tries = total_tries + 1 # reset tries
                                _delay = initial_wait  # reset delay
                                continue
                        msg = str(f'Function: {f.__name__}\n'
                                  f'Failed despite best efforts after {total_tries} tries.\n'
                                  f'args: {print_args}, kwargs: {kwargs}')
                        log(msg, logger)
                        raise
                    msg = str(f'Function: {f.__name__}\n'
                              f'Exception: {e}\n'
                              f'Retrying in {_delay}s args: {print_args}, kwargs: {kwargs}')
                    if not no_print:
                        log(msg, logger)
                    time.sleep(_delay)
                    _delay *= backoff_factor

        return func_with_retries
    return retry_decorator


def log(msg, logger=None):
    if logger:
        logger.warning(msg)
    else:
        print(msg)


def test_func(*args, **kwargs):
    rnd = random.random()
    if rnd < .2:
        raise ConnectionAbortedError('Connection was aborted :(')
    elif rnd < .4:
        raise ConnectionRefusedError('Connection was refused :/')
    elif rnd < .8:
        raise ConnectionResetError('Guess the connection was reset')
    else:
        return 'Yay!!'


# RQ addition.
@retry(FileNotFoundError, total_tries=7, initial_wait=0.5, backoff_factor=1.5)
def wait_on_file(file, pause=0):
    if pause:
        print(f'wait on file pause = {pause}')
        time.sleep(pause)
        pause = 0
    if not os.path.exists(file):
        raise FileNotFoundError(f"{file}")
    return True


if __name__ == '__main__':
    # wrapper = retry((ConnectionAbortedError), tries=3, delay=.2, backoff=1, logger=logger)
    # wrapped_test_func = wrapper(test_func)
    # print(wrapped_test_func('hi', 'bye', hi='ciao'))

    wrapper_all_exceptions = retry(Exception, total_tries=2, logger=logger)
    wrapped_test_func = wrapper_all_exceptions(test_func)
    print(wrapped_test_func('hi', 'bye', hi='ciao'))


# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  __init__.py                                                                          #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

# __init__.py = Ensure the CSI_MODULE environment variable is correctly set to the current module
# name so that module specific initialization done in Common.init.CFG works.

import os

MODULE = "MODULE_NOT_SET"
if __package__:                         # This succeeds if the module is run with 'python -m ./Module_Name'
    MODULE = __package__                # and the __init__.py file is in the directory ./Module_Name
elif __file__.endswith('__init__.py'):  # This should be redundant and unnecessary but...
    MODULE = __file__.split('/')[-2]
else:
    raise RuntimeWarning(f"[{MODULE}/__init__.py] Something went horribly wrong...")

if not os.getenv("CSI_MODULE"):
    os.environ["CSI_MODULE"] = MODULE


from .main import MainHandler

__all__ = ['MainHandler', 'sqs_message']

handler = MainHandler()
new_request = handler.new_request
post_event = handler.post_event
update_response = handler.update_response
sqs_message = handler.sqs_message



# ##################################################################################################
#  Copyright (c) 2024.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  If used, attribution of open-source code is included where required by original author          #
#                                                                                                  #
#  Filename:  setup.py                                                                             #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

from setuptools import setup, find_packages

setup(
    name='csi_genai_lib',
    version='0.1.6',
    description='Caber GenAI Chat Capture Library',
    author='Rob Quiros',
    author_email='dev@caber.com',
    packages=find_packages(),
    # package_data={
    #     'csi_genai_lib': [
    #         '../csiMVP/Common/config.json',
    #         '../csiMVP/Common/init.py',
    #         '../csiMVP/Common/remote_open.py',
    #         '../csiMVP/Common/tf_config_load.py',
    #         '../csiMVP/Toolbox/aws_init.py',
    #         '../csiMVP/Toolbox/goodies.py',
    #         '../csiMVP/Toolbox/json_encoder.py',
    #         '../csiMVP/Toolbox/retry.py',
    #         '../csiMVP/Toolbox/sqs_init.py',
    #         '../csiMVP/Toolbox/supabase.py',
    #     ],
    # },
    install_requires=[
        # 'pandas==1.5.3',
        # 'numpy==1.26.4',
        # 'packaging==23.2',
        'boto3>=1.18.1',
        'smart-open[s3]>=5.1.0',
        'requests>=2.26.0',
        # 'requests_aws4auth>=1.1.1',
        'urllib3>=1.25.11',
        # 'frozendict',
        # 'dnspython==2.4.0',
        # 'netifaces2>=0.0.18',
        # 'python-nmap>=0.7.1'
    ],
)



# ##################################################################################################
#  Copyright (c) 2024.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  streamtap.py                                                                         #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

import os
import uuid
import inspect
import hashlib

from .remote_open import ropen, validate_all_store_configs
from .init import CFG


validated = False
bsfxx = CFG.G["filenameSuffixes"]["body"].strip('.')
bsfxx = f".{bsfxx}" if bsfxx else ""
rsfxx = CFG.G["remoteSuffix"].strip('.')
rsfxx = f".{rsfxx}" if rsfxx else ""


class StreamTap:
    """
    Class initialization inputs:
    instream     (required - file-like object): The originating stream that this class will read

    temp_path    (required - string): local machine directory where the tapped byte stream will be written

    chunk      (optional - int): Number of bytes to yeild. Defaults to maxBodysize in config fiel.  If instream
                is an iterator with a fixed chunk size, chunk must equal that chunk size exactly!

    file_id  (optional - uuid): UUID used to create the file name for the tapped stream.  If not supplied
                 a machine specific uuid will be created.


    Class properties (in addition to base class properties):
    local_file   (string) Path and filename where the copy of the bytes is stored.
    remote_file  (string) Bucket, prefix, and object name of the tapped stream
    """

    def __init__(self, instream, chunk_size=1024, content_length=0, uid=None):
        global validated
        global bsfxx
        global rsfxx

        if not validated:
            validated = validate_all_store_configs(CFG.sspfx)

        self._instrm = instream
        self._bufmax = CFG.me("maxBodyLen", default=(1024 * 1024))
        self.local_name = None
        self.remote_name = None
        self.length = content_length
        self.sha256 = b''
        self._chnksz = chunk_size
        self._start = 0
        if uid is None:
            self.uid = uuid.uuid4()
        else:
            self.uid = uid
        self.result = None
        self._hash = hashlib.sha256()
        self._gen = None
        self._done = False
        self.closed = False
        self._isgen = inspect.isgenerator(instream)
        self._isstr = isinstance(instream, (bytes, bytearray))
        try:
            self._isfile = instream.readable()
        except:
            self._isfile = False

        print(f"[{self.uid[0:4]}{self.uid[-8:]}] Stream source is {type(instream).__qualname__}")

        if not self._isfile and not self._isgen and not self._isstr:
            if isinstance(instream, str):
                try:
                    self._instrm = instream.encode('utf-8')
                    self._isstr = True
                except UnicodeEncodeError:
                    raise TypeError(f"[{self.uid[0:4]}{self.uid[-8:]}] Stream source is a string that could not be encoded to bytes")
            else:
                raise TypeError(f"[{self.uid[0:4]}{self.uid[-8:]}] Stream source is not a file-like, bytes-like, or generator object {type(instream).__qualname__}")
        if not self.closed:
            self._done = False
            self.local_name = f"{CFG.localdir}/{self.uid}{bsfxx}"
            # Remote name in Caber's global shared storage (gss) NO CHARS THAT REQUIRE quote_plus HERE!!!!!!!!!
            self.remote_name = f"{CFG.rem_cori_prefix(module=CFG.M['filePrefix'])}/{self.uid}{bsfxx}{rsfxx}"
            self._ftmp = open(self.local_name, 'wb+')         # ALWAYS LOCAL - BUILT-IN OPEN
        else:
            self._done = True
        if self._isstr:
            print(f'[{self.uid[0:4]}{self.uid[-8:]}] STAP Attached to data of type BYTES chunk size = {self._chnksz}')
        if self._isfile:
            print(f'[{self.uid[0:4]}{self.uid[-8:]}] STAP Attached to data of type FILE chunk size = {self._chnksz}')
        if self._isgen:
            print(f'[{self.uid[0:4]}{self.uid[-8:]}] STAP Attached to data of type GENERATOR chunk size = {self._chnksz}')

    def read(self, chunklen=0):
        chunk = b''
        if self.closed: raise ValueError(f"[{self.uid[0:4]}{self.uid[-8:]}] Attempt to read when stream is closed")
        if not isinstance(chunklen, int):
            raise ValueError(f"[{self.uid[0:4]}{self.uid[-8:]}] Argument 'chunklen' to read(chunklen) must be of type int.")
        if self._chnksz == 0 and chunklen <= 0:   # If a chunk size was not set at init, it can be set by chunklen
            self._chnksz = self._bufmax           #   otherwise chunklen is ignored
        elif self._chnksz == 0:
            self._chnksz = chunklen

        # Enable a bytes-like object to be read like a file
        if self._isstr:
            if not self._done:
                if self._chnksz > len(self._instrm[self._start:]):
                    chunk = self._instrm[self._start:]
                    # Process all  or the last remaining part of instrm
                    self.length += len(chunk)
                    self._ftmp.write(chunk)
                    self._hash.update(chunk)
                    self._done = True
                    print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent final chunk {self.length} bytes tapped")
                else:
                    # Process the next chunk of instream
                    chunk = self._instrm[self._start:self.length]
                    self._ftmp.write(chunk)
                    self._hash.update(chunk)
                    self._start = self.length
                    self.length += len(chunk)
                    # print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent {len(chunk)} byte chunk. ({self.length} bytes so far)")

        # Enable a generator to be read like a file
        elif self._isgen:
            if not self._done:
                try:
                    chunk = next(self._instrm, b'')
                    if chunk == b'':
                        raise StopIteration
                    self._ftmp.write(chunk)
                    self._hash.update(chunk)
                    self.length += len(chunk)
                    # print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent {len(chunk)} byte chunk. ({self.length} bytes so far)")
                except StopIteration:
                    print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent final chunk {self.length} bytes tapped")
                    self._done = True

        # Enable a file to be read like a file :)
        elif self._isfile:
            if not self._done:
                try:
                    chunk = self._instrm.read(self._chnksz)
                    if chunk == b'':
                        raise EOFError
                    self._ftmp.write(chunk)
                    self._hash.update(chunk)
                    self.length += len(chunk)
                    # print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent {len(chunk)} byte chunk. ({self.length} bytes so far)")
                except EOFError:
                    print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP Sent final chunk {self.length} bytes tapped")
                    self._done = True

        else:
            print(f'[{self.uid}][ERROR] How did I get here?  instream type is {type(self._instrm)} done={self._done}')

        if self._done:
            if self._isfile:
                try:
                    if not self._instrm.closed:
                        self._instrm.close()
                except:
                    pass
        # print(f"[{self.uid}][DEBUG] Returning chunk of length={len(chunk)}")
        return chunk

    def close(self, delete_temp=True):
        fsize = 0
        fosize = 0
        if self._ftmp.seekable():
            # print(f"[{self.uid}][DEBUG] Local temp file seekable")
            fsize = self._ftmp.tell()
            fosize = os.stat(self.local_name).st_size
            self._ftmp.seek(0)
        else:
            self._ftmp.close()
            # print(f"[{self.uid}][DEBUG] Local temp file NOT seekable")
            if os.path.exists(self.local_name):
                fsize = os.stat(self.local_name).st_size
                self._ftmp = open(self.local_name, 'rb')         # ALWAYS LOCAL - BUILT-IN OPEN
            else:
                print(f"[{self.uid}][ERROR] What? Local temp file no longer exists! {self.local_name}")
                self.result = b''

        # Get the SHA256 before writing the file so we can use the SHA256 as the filename.  Common to see the same
        # data being read and written many times for the same transaction.
        # TODO: Optimization.  Ideally we would check to see if a file with the same hash exists before writing
        #    to avoid wasted cycles.  Probably best to do the check only for file lengths over a certain size
        #    since checking has a cost too.  Could also use a local lru_cache...
        # TODO: Decompress files. It might also make sense to see if a file is compressed so we don't
        #    don't compress it again when writing it via smart-open.  If we decompress here then
        #    get the sha256 we'll avoid work later on since there won't be duplicate compressed/uncompressed
        #    content to deal with.
        self.sha256 = self._hash.hexdigest()
        self.remote_name = self.remote_name.replace(self.uid, self.sha256)

        if fsize != self.length:
            print(f"[{self.uid}][WARNING] Local temp file length inconsistent {fsize}<>{fosize}<>{self.length}")

        if CFG.G.get("minObjSize", 0) < fsize <= CFG.me("maxInlineBodyLen", 256):
            self.result = self._ftmp.read()
        elif fsize >= CFG.G.get("minObjSize", 0):
            with ropen(self.remote_name, 'wb') as fout:
                print(f"[{self.uid[0:4]}{self.uid[-8:]}] STAP writing {self.length} bytes to remote {self.remote_name}")
                fout.write(self._ftmp.read())
            self.result = f"{self.remote_name}"
        else:
            self.result = b''

        self._ftmp.close()
        if delete_temp:
            os.remove(self.local_name)

        self.closed = True
        self._done = True
        return self.result

    def readable(self): return True
    def writable(self): return False
    def seekable(self): return False

    # The __iter__ and __next__ functions turn the class into a generator even though it uses
    # the read function under the covers
    def __iter__(self):
        # returning __iter__ object
        return self

    def __next__(self):
        while not self._done:
            return self.read(self._chnksz)
        if self._done:
            raise StopIteration

    def __len__(self):
        return self.length

    def __del__(self):
        if not self.closed:
            self.close()

    __call__ = __next__



# ##################################################################################################
#  Copyright (c) 2022.    Caber Systems, Inc.                                                      #
#  All rights reserved.                                                                            #
#                                                                                                  #
#  CABER SYSTEMS CONFIDENTIAL SOURCE CODE                                                          #
#  No license is granted to use, copy, or share this software outside of Caber Systems, Inc.       #
#                                                                                                  #
#  Filename:  json_encoder.py                                                                      #
#  Authors:  Rob Quiros <rob@caber.com>  rlq                                                       #
# ##################################################################################################

import json
import binascii
import base64
from datetime import datetime

try:
    import pandas as pd
except ImportError:
    pd = None


class extEncoder(json.JSONEncoder):
    # Extend the json.JSONEncoder class to support datetime, bytes, and NaN, NaT, None, NA, etc.
    # by overloading the json method default.  This method cannot be used to convert sets to lists,
    # or dataframes to dicts, etc.
    # From the JSON docs:  https://docs.python.org/3/library/json.html
    #    ``default(obj)`` is a function that should return a serializable version
    #     of obj or raise TypeError. The default simply raises TypeError.
    def default(self, obj):
        # Match all the types you want to handle in your converter
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, bytes):
            try:
                obj = obj.decode('utf-8')
                return obj
            except UnicodeDecodeError as err:
                try:
                    obj = base64.b64encode(obj).decode('utf-8')
                    return obj
                except binascii.Error:
                    pass
        elif pd is not None and pd.isna(obj) and not isinstance(obj, (pd.DataFrame, pd.Series)):
            return None
        elif not isinstance(obj, (str, int, float, list, dict, bool)):
            try:
                return json.JSONEncoder.default(self, obj)
            except (ValueError, TypeError):
                return None

        # Call the default method for (str, int, float, list, dict, bool)
        return json.JSONEncoder.default(self, obj)
